"""
The GEOWATCH module

Useful environs:
    DVC_DATA_DPATH=$(geowatch_dvc --tags='phase2_data' --hardware=auto)
    DVC_EXPT_DPATH=$(geowatch_dvc --tags='phase2_expt' --hardware=auto)
"""
import os
import sys
import ubelt as ub
import warnings


__version__ = '0.6.2'


# ~/code/watch/dev/maintain/generate_authors.py
__author__ = 'GEOWATCH Developers, Kitware Inc., Jon Crall, David Joy, Matthew Bernstein, Connor Greenwell, Benjamin Brodie, Peri Akiva, Usman Rafique, Jacob DeRosa, Matthew Purri, Ajay Upadhyaya, Ji Won Suh, Jacob Birge, Ryan LaClair, Scott Workman, Dexter Lau, Sergii Skakun, Aram Ansary Ogholbake, Cohen Archbold, Bane Sullivan, Srikumar Sastry, Armin Hadzic'

__author_email__ = 'kitware@kitware.com, jon.crall@kitware.com'
__url__ = 'https://gitlab.kitware.com/computer-vision/geowatch'


os.environ['USE_PYGEOS'] = '0'


WATCH_PREIMPORT_VARIANTS = {
    'variant1': ['geopandas', 'pyproj', 'gdal'],  # align-crs on horologic
    'variant2': ['pyproj', 'gdal'],               # CI machine
    'variant3': ['geopandas', 'pyproj'],          # delay gdal import
    'none': [],                                   # no pre-imports
    '0': [],                                      # no pre-imports
}


def _handle_hidden_commands():
    """
    Hidden developer features
    """
    if ub.argflag('--warntb') or os.environ.get('WARN_WITH_TRACEBACK', ''):
        import xdev
        xdev.make_warnings_print_tracebacks()


def _import_troublesome_module(modname):
    """
    Defines exactly how to import each troublesome (binary) module
    """
    if modname == 'gdal':
        from osgeo import gdal as module
        module.UseExceptions()
    elif modname == 'pyproj':
        import pyproj as module
        from pyproj import CRS  # NOQA
    elif modname == 'geopandas':
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', (
                '.*is incompatible with the GEOS version '
                'PyGEOS was compiled with.*'))
            import geopandas as module
    elif modname == 'rasterio':
        import rasterio as module
    elif modname == 'fiona':
        import fiona as module
    # elif modname == 'pygeos':
    #     import pygeos as module
    elif modname == 'torch':
        import torch as module
    elif modname == 'numpy':
        import numpy as module
    else:
        raise KeyError(modname)
    return module


def _is_running_a_fast_cli_tool():
    """
    Determine if we are running a fast cli tool.

    This is used to short circuit the pre-imports for certain command line
    tools so we don't incur the import time, which can be multiple seconds.

    TODO:
        Is there a better way to prevent certain entry points from executing
        the pre-imports?

    Notes:
        There is crazy sys.argv behavior with -m [SO42076706]_.

    References:
        .. [SO42076706] https://stackoverflow.com/questions/42076706/sys-argv-behavior-with-python-m
    """
    if os.environ.get('_ARGCOMPLETE', ''):
        return True
    if sys.argv and 'geowatch_dvc' in sys.argv[0] or 'geowatch_dvc' in sys.argv[0]:
        return True
    if sys.argv and len(sys.argv) == 1:
        # No args given case
        return True
    if sys.argv and sys.argv == ['-m']:
        return True
    if sys.argv and ('--help' in sys.argv or '-h' in sys.argv):
        return True
    return False


def _execute_ordered_preimports():
    """
    The order in which certain modules with binary libraries are imported can
    impact runtime stability and can even cause crashes if a specific order
    isnt used.

    There are several known good configurations registered in the
    ``WATCH_PREIMPORT_VARIANTS`` dictionary and the setting is
    controlled by the ``WATCH_PREIMPORT`` environment variable.
    """

    # Shorter alias because we are using it now
    WATCH_PREIMPORT = os.environ.get('WATCH_PREIMPORT', 'auto')
    WATCH_PREIMPORT = os.environ.get('WATCH_HACK_IMPORT_ORDER', WATCH_PREIMPORT)

    if not WATCH_PREIMPORT:
        return

    if WATCH_PREIMPORT == 'auto':
        if _is_running_a_fast_cli_tool():
            watch_preimport = None
        else:
            # This is the "known" best order for importing
            watch_preimport = WATCH_PREIMPORT_VARIANTS['variant1']
    elif WATCH_PREIMPORT in WATCH_PREIMPORT_VARIANTS:
        watch_preimport = WATCH_PREIMPORT_VARIANTS[WATCH_PREIMPORT]
    elif WATCH_PREIMPORT.lower() in {'0', 'false', 'no', ''}:
        watch_preimport = None
    else:
        watch_preimport = WATCH_PREIMPORT.split(',')

    if watch_preimport is not None:
        for modname in watch_preimport:
            _import_troublesome_module(modname)


_handle_hidden_commands()
_execute_ordered_preimports()


if 0:
    # Disable
    from watch.monkey import monkey_numpy  # NOQA
    monkey_numpy.patch_numpy_dtypes()


if 'hard-to-inspect-key' in vars():
    # Defined to hack jon's editor into autocompleting this.
    find_dvc_dpath = None


__devnotes__ = """

# Command to autogenerate lazy imports for this file
mkinit -m watch --lazy_loader --diff
mkinit -m watch --lazy_loader -w

# Debug import time
python -X importtime -c "import watch"
WATCH_HACK_IMPORT_ORDER=variant3 python -X importtime -c "import watch"
WATCH_HACK_IMPORT_ORDER=variant1 python -X importtime -c "import watch"
"""


# Choose which submodules (and which submodule attributes) to expose
__submodules__ = {
    '*': [],  # include all modules, but don't expose attributes
    'demo': ['coerce_kwcoco'],
    'utils': ['find_smart_dvc_dpath', 'find_dvc_dpath']
}


# --- AUTOGENERATED LAZY IMPORTS ---

import lazy_loader  # NOQA

__getattr__, __dir__, __all__ = lazy_loader.attach(
    __name__,
    submodules={
        'cli',
        'demo',
        'exceptions',
        'gis',
        'heuristics',
        'mlops',
        'monkey',
        'rc',
        'stac',
        'tasks',
        'utils',
    },
    submod_attrs={
        'demo': [
            'coerce_kwcoco',
        ],
        'utils': [
            'find_smart_dvc_dpath',
            'find_dvc_dpath',
        ],
    },
)

__all__ = ['WATCH_PREIMPORT_VARIANTS', 'cli', 'coerce_kwcoco', 'demo',
           'exceptions', 'find_dvc_dpath', 'find_smart_dvc_dpath', 'gis',
           'heuristics', 'mlops', 'monkey', 'rc', 'stac', 'tasks', 'utils']



import os
import shutil

import torch
import numpy as np
from PIL import Image
import matplotlib.pyplot as plt
from watch.tasks.rutgers_material_change_detection.utils.util_misc import create_gif


class VideoPlotMaker:
    def __init__(self, save_dir, task_mode, rgb_func, temp_dir="./temp/"):
        self.save_dir = save_dir
        self.rgb_func = rgb_func

        # Get plot function.
        plot_methods = {"total_bin_change": self.plot_total_bin_change}
        self.plot_function = plot_methods[task_mode]

        # Create temporary directory.
        if os.path.isdir(temp_dir):
            shutil.rmtree(temp_dir)
        os.makedirs(temp_dir, exist_ok=True)

        self.temp_dir = temp_dir

    def __call__(self, video, target, prediction, n_frames, datetimes, save_name):
        # Create and save individual plots.
        save_paths = self.plot_function(video, target, prediction, n_frames, datetimes)

        # Generate GIF of plots.
        ## Load images with PIL
        np_images = [np.asarray(Image.open(p)) for p in save_paths]
        save_path = os.path.join(self.save_dir, save_name + ".gif")

        ## Create GIF
        create_gif(np_images, save_path)

    def plot_total_bin_change(self, video, target, prediction, n_frames, datetimes):

        # Get ground truth image.
        ground_truth = np.clip(target, 0, 1)

        # Get colored prediction.
        model_class_pred = np.argmax(prediction, axis=0)
        pred_canvas = np.zeros((model_class_pred.shape[0], model_class_pred.shape[1], 3), dtype="uint8")

        # Get True Positives.
        x, y = np.where((model_class_pred == 1) & (target == 1))
        true_positive_color = [255, 255, 255]
        pred_canvas[x, y, :] = np.repeat(np.array(true_positive_color)[np.newaxis], repeats=x.shape[0], axis=0)

        # Get False Positives.
        x, y = np.where((model_class_pred == 1) & (target == 0))
        false_positive_color = [0, 255, 255]
        pred_canvas[x, y, :] = np.repeat(np.array(false_positive_color)[np.newaxis], repeats=x.shape[0], axis=0)

        # Get False Negatives.
        x, y = np.where((model_class_pred == 0) & (target == 1))
        false_negative_color = [255, 0, 0]
        pred_canvas[x, y, :] = np.repeat(np.array(false_negative_color)[np.newaxis], repeats=x.shape[0], axis=0)

        # Compute confidence mask from model prediction.
        conf_mask = torch.softmax(torch.from_numpy(prediction), dim=0)[1]

        save_paths = []
        for frame_index in range(n_frames + 1):
            # Get RGB frame.
            if frame_index == n_frames:
                # Add blank frame.
                rgb_frame = np.zeros([target.shape[0], target.shape[1], 3])
            else:
                rgb_frame = self.rgb_func(video[frame_index])

                # Get datetime.
                datetime_str = datetimes[frame_index].strftime("%m/%d/%Y")

            _, axes = plt.subplots(2, 2, figsize=(5, 5), dpi=300)  # RGB, Conf, Pred, GT
            axes[0][0].imshow(rgb_frame)
            if frame_index == n_frames:
                axes[0][0].set_title("")
            else:
                axes[0][0].set_title(f"[{frame_index+1}/{n_frames}] - " + datetime_str)
            axes[0][0].axis("off")

            axes[0][1].imshow(conf_mask, "viridis", vmin=0, vmax=1)
            axes[0][1].set_title("Confidence")
            axes[0][1].axis("off")

            axes[1][0].imshow(pred_canvas, cmap=None)
            axes[1][0].set_title("Prediction")
            axes[1][0].axis("off")
            axes[1][1].imshow(ground_truth, cmap="gray")
            axes[1][1].set_title("Ground Truth")
            axes[1][1].axis("off")
            plt.tight_layout()

            # Save figure.
            save_path = os.path.join(self.temp_dir, str(frame_index).zfill(5) + ".png")
            plt.savefig(save_path, dpi=300, pad_inches=0.0)
            save_paths.append(save_path)
            plt.close()
        return save_paths


def get_instance_visualization(
    task_mode,
    model_pred,
    video,
    target,
    rgb_func,
    last_frame_index=-1,
    unnorm_func=None,
    save_path=None,
    title_str=None,
):
    """Generate a plot to visualize an instance of the model prediction, input, and ground truth.

    Args:
        task_mode (str): [description]
        model_pred (?): The models prediction will depend on the task_mode.
        video (np.array): A float numpy array of shape [n_frames, n_channels, height, width].
        target (?): Depending on the task_mode parameter, this can be a np.array of shape [n_classes, height, width], integer, etc.
        rgb_func (?): A function to translate a multi-modal image into a uint8 numpy array ready for plotting.
        last_frame_index (int, optional): An integer representing the last frame of video that contains real data instead of padded data.
        unnorm_func (?, optional): A function to unnormalize a normalized video. Defaults to None.
        save_path (str, optional): A path to save the figure to. Defaults to None.

    Raises:
        NotImplementedError: [description]

    Returns:
        plt.Figure: A subplot figure.
    """

    # Undo the normalization done to the image.
    if unnorm_func is not None:
        video = unnorm_func(video)

    # Get and format ground truth and prediction representations.
    if task_mode == "total_bin_change":
        # Format input images to RGB form.
        first_frame_rgb = rgb_func(video[0])
        last_frame_rgb = rgb_func(video[last_frame_index])

        # Format ground truth image.
        target = np.clip(target, 0, 1).astype("float32")

        # Format prediction and confidence mask.
        model_class_pred = np.argmax(model_pred, axis=0)
        pred_canvas = np.zeros((model_class_pred.shape[0], model_class_pred.shape[1], 3), dtype="uint8")

        # Get True Positives.
        x, y = np.where((model_class_pred == 1) & (target == 1))
        true_positive_color = [255, 255, 255]
        pred_canvas[x, y, :] = np.repeat(np.array(true_positive_color)[np.newaxis], repeats=x.shape[0], axis=0)

        # Get False Positives.
        x, y = np.where((model_class_pred == 1) & (target == 0))
        false_positive_color = [0, 255, 255]
        pred_canvas[x, y, :] = np.repeat(np.array(false_positive_color)[np.newaxis], repeats=x.shape[0], axis=0)

        # Get False Negatives.
        x, y = np.where((model_class_pred == 0) & (target == 1))
        false_negative_color = [255, 0, 0]
        pred_canvas[x, y, :] = np.repeat(np.array(false_negative_color)[np.newaxis], repeats=x.shape[0], axis=0)

        # Compute confidence mask from model prediction.
        conf_mask = torch.softmax(torch.from_numpy(model_pred))[1]

        # Create plot to store visualization of inputs and prediction.
        fig, axes = plt.subplots(1, 5)
        axes[0].imshow(first_frame_rgb)  # First frame of video (RGB).
        axes[0].set_title("First Frame")
        axes[0].axis("off")
        axes[1].imshow(last_frame_rgb)  # Last frame of video (RGB).
        axes[1].set_title("Last Frame")
        axes[1].axis("off")
        axes[2].imshow(pred_canvas, cmap=None)  # Prediction.
        axes[2].set_title("Prediction")
        axes[2].axis("off")
        axes[3].imshow(conf_mask, "viridis", vmin=0, vmax=1)  # Confidence of change.
        axes[3].set_title("Change Conf")
        axes[3].axis("off")
        axes[4].imshow(target, "gray")  # Ground truth.
        axes[4].set_title("Ground Truth")
        axes[4].axis("off")
        plt.tight_layout()
    elif task_mode == "ss_mat_recon":
        breakpoint()
        pass
        # # Show the first and last frame inputs as well as the predicted reconstruction.
        # # Format input images to RGB form.
        # n_frames = int((data['active_frames'][batch_index].sum() - 1).item())

        # # Create plot to store visualization of inputs and prediction.
        # if self.cfg.loss.discretize:
        #     fig, axes = plt.subplots(3, n_frames)
        #     in_frames = dataset.unnormalize(data['video'][batch_index].detach().cpu())
        #     pred_frames = dataset.unnormalize(model_output[task_mode][batch_index].detach().cpu())

        #     _, cluster_ids = torch.max(model_output['mat_cluster_ids']['layer1'][:, batch_index], dim=-1)
        #     cluster_ids = cluster_ids.detach().cpu().numpy()

        #     color_code = self.model.color_codes['layer1']

        #     material_cluster_ids = np.take(color_code, cluster_ids, axis=0)

        #     for i in range(n_frames):
        #         axes[0][i].imshow(dataset.to_rgb(in_frames[i]))  # RGB input.
        #         axes[0][i].axis('off')

        #         axes[1][i].imshow(dataset.to_rgb(pred_frames[i]))  # RGB prediction.
        #         axes[1][i].axis('off')

        #         axes[2][i].imshow(material_cluster_ids[i])  # RGB prediction.
        #         axes[2][i].axis('off')

        # else:
        #     fig, axes = plt.subplots(2, n_frames)
        #     in_frames = dataset.unnormalize(data['video'][batch_index].detach().cpu())
        #     pred_frames = dataset.unnormalize(model_output[task_mode][batch_index].detach().cpu())

        #     for i in range(n_frames):
        #         axes[0][i].imshow(dataset.to_rgb(in_frames[i]))  # RGB input.
        #         axes[0][i].axis('off')

        #         axes[1][i].imshow(dataset.to_rgb(pred_frames[i]))  # RGB prediction.
        #         axes[1][i].axis('off')

        # plt.tight_layout()
    elif task_mode in ["ss_arrow_of_time", "ss_splice_change"]:
        # TODO: Implement visualization for this method.
        fig, axes = plt.subplots(1, 5)
    else:
        raise NotImplementedError(f'Target visualization for mode "{task_mode}" not implemented.')

    if save_path is not None:
        plt.savefig(save_path, dpi=300, pad_inches=0.0)
    else:
        return fig, axes



import os
import argparse
from glob import glob

import torch
import kwimage
import numpy as np
import ubelt as ub
from tqdm import tqdm
from osgeo import gdal


from watch.tasks.rutgers_material_change_detection.models import build_model
from watch.tasks.rutgers_material_change_detection.datasets import build_dataset, create_loader
from watch.tasks.rutgers_material_change_detection.utils.util_misc import get_n_frames, generate_video_slice_object


class Evaluator:
    def __init__(self, args):
        model_path = args.model_path
        kwcoco_file_dir = args.input_kwcoco_dir
        self.save_kwcoco_path = args.output_kwcoco_file_path
        self.heatmap_pred_channel_names = args.heatmap_pred_channel_names

        # Get model configuration file.
        if os.path.isfile(model_path) is False:
            raise FileNotFoundError(f"Model not found at: {model_path}")
        model_data = torch.load(model_path)
        self.cfg = model_data["cfg"]

        # Update config based on command line arguments.
        if args.batch_size != -1:
            self.cfg.batch_size = args.batch_size
        if args.stride is not None:
            self.cfg.stride = args.stride

        # Get device.
        self.device = args.device

        # Build dataset object.
        self.eval_dataset, video_slice = self._build_dataset(kwcoco_file_dir)

        # Create dataloader.
        self.eval_loader = create_loader(self.eval_dataset, args.split, self.cfg.batch_size, args.n_workers)

        # Build model.
        self.model = self._build_model(model_data, video_slice)

    def _build_dataset(self, kwcoco_file_dir):
        # Build dataset object.
        n_frames = get_n_frames(self.cfg.dataset.n_frames, self.cfg.task_mode)
        video_slice = generate_video_slice_object(
            height=self.cfg.height,
            width=self.cfg.width,
            n_frames=n_frames,
            scale=self.cfg.scale,
            stride=self.cfg.stride,
        )

        dataset = build_dataset(
            self.cfg.dataset.name,
            args.split,
            video_slice,
            self.cfg.task_mode,
            seed_num=self.cfg.seed_num,
            normalize_mode=self.cfg.normalize_mode,
            channels=self.cfg.dataset.channels,
            max_iterations=None,
            overwrite_dset_dir=kwcoco_file_dir,
        )
        return dataset, video_slice

    def _build_model(self, model_data, video_slice):
        # TODO: Check that model is trained for this dataset.
        model = build_model(
            self.cfg,
            video_slice,
            self.eval_dataset.n_channels,
            self.eval_dataset.max_frames,
            self.eval_dataset.n_classes,
            device=self.device,
        ).to(self.device)

        # Update model weights from checkpoint.
        model.load_state_dict(model_data["state_dict"], strict=True)
        model.to(self.device)
        model.eval()

        return model

    def gen_kwcoco_file(self):
        self.model.eval()

        scale = self.cfg.scale

        # TODO: Make this run in parallel with dataloader.

        # Create a random subset of examples from dataloader to choose from.
        region_images, region_image_ids, overlap_masks = {}, {}, {}
        for examples in tqdm(self.eval_loader):
            # Load videos onto GPU memory.
            examples["video"] = examples["video"].to(self.device, non_blocking=True)

            # Get model prediction.
            model_outputs = self.model(examples)
            B = examples["video"].shape[0]
            for b in range(B):
                # Create region canvas or paste predictions.
                region_name = examples["region_name"][b]
                if region_name not in list(region_images.keys()):
                    og_height, og_width = examples["crop_info"][b]["og_height"], examples["crop_info"][b]["og_width"]
                    # Create canvases.
                    canvases = {
                        "conf": np.zeros([og_height, og_width, 2], dtype="float"),
                    }
                    region_images[region_name] = canvases
                    overlap_masks[region_name] = np.zeros([og_height, og_width], dtype="int")

                # Format pred/target.
                H0, W0, dH, dW = examples["crop_info"][b]["space_crop_slice"]
                model_pred = model_outputs[self.cfg.task_mode][b].detach().cpu()
                # NOTE: CONFIDENCE PREDICTION IS COMPUTED AS PROBS INSTEAD OF PREDS
                conf = torch.softmax(model_pred, dim=0).numpy()[:, ::scale, ::scale][:, :dH, :dW].transpose(1, 2, 0)

                # Paste crop pred/targets into canvases.
                region_images[region_name]["conf"][H0 : H0 + dH, W0 : W0 + dW] += conf
                overlap_masks[region_name][H0 : H0 + dH, W0 : W0 + dW] += 1

                # Get gids.
                if not (region_name in region_image_ids.keys()):
                    image_ids = self.eval_dataset.video_dataset[examples["crop_info"][b]["video_id"]]
                    region_image_ids[region_name] = image_ids

        # Get kwcoco dataset file from Dataset class.
        kwcoco_dataset = self.eval_dataset.coco_dset.copy()
        kwcoco_dataset.clear_annotations()
        kwcoco_dataset.reroot(absolute=True)  # Make all paths absolute
        kwcoco_dataset.fpath = self.save_kwcoco_path  # Change output file path and bundle path
        kwcoco_dataset.reroot(absolute=False)  # Reroot in the new bundle path

        # Set path to save auxiliary files.
        bundle_dpath = ub.Path(kwcoco_dataset.bundle_dpath)

        # Save full region images.
        for region_name, canvases in tqdm(region_images.items()):
            conf_image = canvases["conf"]
            image_ids = region_image_ids[region_name]

            # Normalize heatmap confidence image by number of predictions made per pixel.
            conf_image = conf_image / overlap_masks[region_name][:, :, None]

            # Write image to disk.
            ## Try only saving one image instead of many copies.
            region_conf_pred_save_dir = os.path.join(bundle_dpath, "aux_" + region_name)
            os.makedirs(region_conf_pred_save_dir, exist_ok=True)
            region_conf_pred_save_path = os.path.join(region_conf_pred_save_dir, "conf_pred.tif")

            # Get geo metadata
            # TODO: Update this method.
            gid = image_ids[1]
            coco_img = kwcoco_dataset._coco_image(gid)
            dset_root = coco_img.dset.bundle_dpath
            image_folder_name = kwcoco_dataset.index.imgs[gid]["name"]
            long_region_name = kwcoco_dataset._coco_image(gid).img["parent_name"]
            long_region_name = "_".join(long_region_name.split("_")[:3])

            try:
                image_dir = os.path.join(dset_root, "KR_R001", "S2", "affine_warp", image_folder_name)
                image_paths = sorted(glob(image_dir + "/*.tif"))
                ds = gdal.Open(image_paths[5])  # Get blue band
            except IndexError:
                image_dir = os.path.join(dset_root, "KR_R002", "S2", "affine_warp", image_folder_name)
                image_paths = sorted(glob(image_dir + "/*.tif"))
                ds = gdal.Open(image_paths[5])  # Get blue band

            ## Save image using gdal.
            driver = gdal.GetDriverByName("GTiff")
            outdata = driver.Create(
                region_conf_pred_save_path, conf_image.shape[1], conf_image.shape[0], 2, gdal.GDT_Float64
            )

            non_sal_band = outdata.GetRasterBand(1)
            non_sal_band.WriteArray(conf_image[:, :, 0])
            sal_band = outdata.GetRasterBand(2)
            sal_band.WriteArray(conf_image[:, :, 1])
            outdata.SetGeoTransform(ds.GetGeoTransform())
            outdata.SetProjection(ds.GetProjection())
            outdata.FlushCache()
            outdata = None

            for image_id in image_ids:
                img = kwcoco_dataset.index.imgs[image_id]

                vid_from_img = kwimage.Affine.coerce(img["warp_img_to_vid"])
                img_from_vid = vid_from_img.inv()

                img.get("auxiliary", []).append(
                    {
                        "file_name": region_conf_pred_save_path,
                        "channels": self.heatmap_pred_channel_names,
                        "height": conf_image.shape[0],
                        "width": conf_image.shape[1],
                        "num_bands": 2,
                        "warp_aux_to_img": img_from_vid.concise(),
                    }
                )
                kwcoco_dataset.index.imgs[image_id] = img

        kwcoco_dataset.dump(self.save_kwcoco_path)
        print(f"Output predictions path: {self.save_kwcoco_path}")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("model_path", type=str, help="Path to the .pth.tar file ")
    parser.add_argument(
        "input_kwcoco_dir",
        type=str,
        help="Path to a directory that contains a vali.kwcoco.json file to get input data from.",
    )
    parser.add_argument(
        "output_kwcoco_file_path", type=str, help="Output predictions placed in a .kwcoco.json at this path."
    )
    parser.add_argument(
        "--device",
        type=str,
        default="cuda",
        choices=["cuda", "cpu"],
        help="The type of hardware to process data with model.",
    )
    parser.add_argument("--n_workers", type=int, default=4, help="Number of CPU processes to load data into model.")
    parser.add_argument("--batch_size", type=int, default=8)
    parser.add_argument("--split", default="valid", help="TODO: Remove this")
    parser.add_argument("--stride", type=int, default=None, help="Set the step that crops are created.")
    parser.add_argument(
        "--heatmap_pred_channel_names",
        type=str,
        default="not_salient|salient",
        help="Overwrite the name of heatmap predictions in the produced kwcoco file.",
    )
    # parser.add_argument(
    #     "--ignore_material_features",
    #     default=False,
    #     action="store_true",
    #     help="Do not add material features to output kwcoco file.",
    # )
    # parser.add_argument(
    #     "--ignore_heatmaps",
    #     default=False,
    #     action="store_true",
    #     help="Do not add confidence heatmaps to output kwcoco file.",
    # )
    # parser.add_argument(
    #     "--ignore_material_change_mask",
    #     default=False,
    #     action="store_true",
    #     help="Do not add material change mask to output kwcoco file.",
    # )
    args = parser.parse_args()

    evaluator = Evaluator(args)
    evaluator.gen_kwcoco_file()



import torch.nn as nn
import torch
from torch.nn import functional as F


class Conv2d(nn.Conv2d):

    def __init__(self, in_channels, out_channels, kernel_size, stride=1,
                 padding=0, dilation=1, groups=1, bias=True):
        super(Conv2d, self).__init__(in_channels, out_channels, kernel_size, stride,
                                     padding, dilation, groups, bias)

    def forward(self, x):
        # return super(Conv2d, self).forward(x)
        weight = self.weight
        weight_mean = weight.mean(dim=(1, 2, 3), keepdim=True)
        weight = weight - weight_mean
        std = weight.view(weight.size(0), -1).std(dim=1).view(-1, 1, 1, 1) + 1e-5
        weight = weight / std.expand_as(weight)
        return F.conv2d(x, weight, self.bias, self.stride,
                        self.padding, self.dilation, self.groups)


class DoubleConv(nn.Module):
    """(convolution => [BN] => ReLU) * 2"""

    def __init__(self, in_channels, out_channels):
        super().__init__()
        self.double_conv = nn.Sequential(
            Conv2d(in_channels, out_channels, kernel_size=3, stride=1, padding=1),
            nn.GroupNorm(32, out_channels),
            nn.LeakyReLU(0.2),
            nn.Dropout(0.5),
            Conv2d(out_channels, out_channels, kernel_size=3, stride=1, padding=1),
            nn.GroupNorm(32, out_channels),
            nn.LeakyReLU(0.2),
        )

    def forward(self, x):
        return self.double_conv(x)


class Down(nn.Module):
    """Downscaling with maxpool then double conv"""

    def __init__(self, in_channels, out_channels):
        super().__init__()
        self.maxpool_conv = nn.Sequential(
            nn.MaxPool2d(2),
            DoubleConv(in_channels, out_channels)
        )

    def forward(self, x):
        return self.maxpool_conv(x)


class Up(nn.Module):
    """Upscaling then double conv"""

    def __init__(self, in_channels, out_channels, bilinear=True):
        super().__init__()

        # if bilinear, use the normal convolutions to reduce the number of
        # channels
        if bilinear:
            self.up = nn.Upsample(scale_factor=2, mode='bilinear', align_corners=True)
        else:
            self.up = nn.ConvTranspose2d(in_channels // 2, in_channels // 2, kernel_size=2, stride=2)

        self.conv = DoubleConv(in_channels, out_channels)
        # Given transposed=1, weight of size [48, 48, 2, 2], 48 -> 32+64//2, instead,
        # expected input[4, 64, 128, 128] to have 48 channels, but got 64
        # channels instead

    def forward(self, x1, x2):
        # print(x1.shape)
        x1 = self.up(x1)
        # input is CHW
        diffY = x2.size()[2] - x1.size()[2]
        diffX = x2.size()[3] - x1.size()[3]

        x1 = F.pad(x1, [diffX // 2, diffX - diffX // 2,
                        diffY // 2, diffY - diffY // 2])

        x = torch.cat([x2, x1], dim=1)
        return self.conv(x)


class OutConv(nn.Module):
    def __init__(self, in_channels, out_channels):
        super(OutConv, self).__init__()
        self.conv = nn.Conv2d(in_channels, out_channels, kernel_size=1)
        # self.conv = nn.Sequential(
        #                             nn.Conv2d(in_channels,out_channels,kernel_size=1),
        #                             nn.Sigmoid()
        #                             )

    def forward(self, x):
        return self.conv(x)


class ShallowSeg(nn.Module):
    def __init__(self, num_channels=3, num_classes=3,
                 bilinear=True, pretrained=False,
                 beta=False, weight_std=False,
                 num_groups=32, out_dim=128, feats=[64, 64, 128, 256, 512]):
        super(ShallowSeg, self).__init__()
        self.num_channels = num_channels
        self.num_classes = num_classes
        self.bilinear = bilinear
        self.out_dim = out_dim
        # feats = [64, 64, 128, 256, 512]
        # feats = [32, 32, 64, 64, 128]

        self.inc = DoubleConv(num_channels, feats[0])
        self.down1 = Down(feats[0], feats[1])
        self.down2 = Down(feats[1], feats[2])
        self.down3 = Down(feats[2], feats[3])
        self.down4 = Down(feats[3], feats[4])

        self.up1 = Up(feats[4] + feats[3], feats[3], bilinear)
        self.up2 = Up(feats[2] + feats[3], feats[2], bilinear)
        self.up3 = Up(feats[1] + feats[2], feats[1], bilinear)
        self.up4 = Up(feats[0] + feats[1], feats[0], bilinear)

        self.outc = OutConv(feats[0], num_classes)
        self.features_outc = OutConv(feats[0], out_dim)

    def forward(self, x):
        # b, c, h, w = x.shape
        # print(x.shape)
        x1 = self.inc(x)
        x2 = self.down1(x1)
        x3 = self.down2(x2)
        x4 = self.down3(x3)
        x5 = self.down4(x4)
        x = self.up1(x5, x4)
        x = self.up2(x, x3)
        x = self.up3(x, x2)
        x = self.up4(x, x1)
        # features = self.features_outc(x)
        # print(dictionary.shape)
        logits = self.outc(x)
        return logits, x5



"""
Define the individual nodes that can be composed in a SMART pipeline.

The topology of the pipeline will define the resulting filesystem structure
used to store results.


CommandLine:
    xdoctest -m watch.mlops.smart_pipeline __doc__:0
    WATCH_DEVCHECK=1 xdoctest -m watch.mlops.smart_pipeline __doc__:1

Example:
    >>> # xdoctest: +SKIP
    >>> from watch.mlops.smart_pipeline import *  # NOQA
    >>> from cmd_queue.util import util_networkx
    >>> #
    >>> config = {
    >>>     'bas_pxl.package_fpath': '/global/models/bas_model2.pt',
    >>>     'bas_pxl.num_workers': 3,
    >>>     'bas_pxl.tta_time': 1,
    >>>     'bas_pxl.test_dataset': '/global/datasets/foobar.kwcoco.zip',
    >>> #
    >>>     'bas_poly.thresh': 0.1,
    >>>     'bas_poly.moving_window_size': 0.1,
    >>> #
    >>>     'sc_pxl.package_fpath': '/global/models/sc_model2.pt',
    >>>     'sc_pxl.tta_fliprot': 8,
    >>>     'sc_poly.use_viterbi': 0,
    >>> }
    >>> #
    >>> dag = make_smart_pipeline('joint_bas_sc')
    >>> # dag = make_smart_pipeline('bas')
    >>> # dag = make_smart_pipeline('sc')
    >>> dag.configure(config, root_dpath='/dag-root/dag-id')
    >>> #
    >>> for node in dag.nodes.values():
    >>>     print('---')
    >>>     print(f'node={type(node)}')
    >>>     print(f'{node.name=}')
    >>>     print(f'{node.config=}')
    >>>     print(f'{node.in_paths=}')
    >>>     print(f'{node.out_paths=}')
    >>>     print(f'{node.resources=}')
    >>>     print(f'{node.algo_params=}')
    >>>     print('node.depends = {}'.format(ub.urepr(node.depends, nl=1, sort=0)))
    >>>     final = node._finalize_templates()
    >>>     print('final = {}'.format(ub.urepr(final, nl=2)))
    >>>     print('---')
    >>> dag.print_graphs()
    >>> print('dag.config = {}'.format(ub.urepr(dag.config, nl=1)))
    >>> dag_templates = {}
    >>> dag_paths = {}
    >>> for node in dag.nodes.values():
    >>>     dag_templates[node.name] = node._build_templates()['node_dpath']
    >>>     dag_paths[node.name] = node._finalize_templates()['node_dpath']
    >>>     print(node.command())
    >>> import rich
    >>> rich.print('dag_templates = {}'.format(
    >>>     ub.urepr(dag_templates, nl=1, sv=1, align=':', sort=0)))
    >>> rich.print('dag_paths = {}'.format(
    >>>     ub.urepr(dag_paths, nl=1, sv=1, align=':', sort=0)))


Example:
    >>> # xdoctest: +REQUIRES(env:WATCH_DEVCHECK)
    >>> from watch.mlops.smart_pipeline import *  # NOQA
    >>> import watch
    >>> expt_dvc_dpath = watch.find_dvc_dpath(tags='phase2_expt', hardware='auto')
    >>> data_dvc_dpath = watch.find_dvc_dpath(tags='phase2_data', hardware='auto')
    >>> #
    >>> config = {}
    >>> config['bas_pxl.test_dataset'] = data_dvc_dpath / 'Drop4-BAS/KR_R001.kwcoco.zip'
    >>> #config['bas_pxl.test_dataset'] = data_dvc_dpath / 'Drop4-BAS/KR_R002.kwcoco.zip'
    >>> #config['bas_pxl.test_dataset'] = data_dvc_dpath / 'Drop4-BAS/BR_R001.kwcoco.zip'
    >>> config['bas_pxl.package_fpath'] = expt_dvc_dpath / 'models/fusion/Drop4-BAS/packages/Drop4_TuneV323_BAS_30GSD_BGRNSH_V2/package_epoch0_step41.pt.pt'
    >>> config['bas_pxl.num_workers'] = 6
    >>> #config['bas_pxl.chip_dims'] = "512,512"
    >>> #config['bas_pxl.time_span'] = "1m"
    >>> #config['bas_pxl.time_sampling'] = "hardish2"
    >>> #config['bas_pxl.use_cloudmask'] = 0
    >>> #config['bas_pxl.set_cover_algo'] = 'approx'
    >>> #config['bas_pxl.resample_invalid_frames'] = 0
    >>> config['bas_poly.thresh'] = 0.1
    >>> #config['sc_pxl.chip_dims'] = "256,256"
    >>> #config['sc_pxl.use_cloudmask'] = 0
    >>> #config['sc_pxl.set_cover_algo'] = 'approx'
    >>> #config['sc_pxl.resample_invalid_frames'] = 0
    >>> config['sc_pxl.num_workers'] = 6
    >>> config['sc_pxl.package_fpath'] = expt_dvc_dpath / 'models/fusion/Drop4-SC/packages/Drop4_tune_V30_8GSD_V3/Drop4_tune_V30_8GSD_V3_epoch=2-step=17334.pt.pt'
    >>> #
    >>> root_dpath = data_dvc_dpath / '_testdag'
    >>> #
    >>> nodes = make_smart_pipeline_nodes()
    >>> #nodes = bas_nodes()
    >>> from watch.mlops.pipeline_nodes import PipelineDAG
    >>> self = dag = PipelineDAG(nodes)
    >>> dag.configure(config=config, root_dpath=root_dpath)
    >>> dag.print_graphs()
    >>> cmd_queue = dag.submit_jobs()
    >>> cmd_queue.write_network_text()
    >>> cmd_queue.rprint()
    >>> #cmd_queue.run()

"""
import ubelt as ub
import shlex
import json
from watch.mlops.pipeline_nodes import ProcessNode


try:
    from xdev import profile  # NOQA
except ImportError:
    profile = ub.identity

PREDICT_NAME  = 'pred'
EVALUATE_NAME = 'eval'


class FeatureComputation(ProcessNode):
    executable = 'python -m watch.cli.run_metrics_framework'
    group_dname = PREDICT_NAME

    # node_dname = 'feats/{src_dset}'

    in_paths = {'src'}

    @profile
    def command(self):
        command = ub.codeblock(
            r'''
            smartwatch teamfeat invariant # TODO
            ''')
        return command

    @property
    def condensed(self):
        condensed = super().condensed
        condensed['src_dset'] = 'todo'
        return condensed


class FeatureUnion(ProcessNode):
    name = 'featunion'
    executable = 'smartwatch feature_union'
    group_dname = PREDICT_NAME
    in_paths = {'src'}
    out_paths = {
        'dst': 'combo_{featunion_id}.kwcoco.zip'
    }

    @profile
    def command(self):
        command = ub.codeblock(
            r'''
            kwcoco union todo
            ''')
        return command


class HeatmapPrediction(ProcessNode):
    executable = 'python -m watch.tasks.fusion.predict'
    group_dname = PREDICT_NAME

    resources = {
        'cpus': 2,
        'gpus': 1,
    }

    perf_params = {
        'num_workers': 2,
        'devices': '0,',
        #'accelerator': 'gpu',
        'batch_size': 1,
        'with_saliency': 'auto',
        'with_class': 'auto',
        'with_change': 'auto',
    }

    in_paths = {
        'package_fpath',
        'test_dataset',
    }

    algo_params = {
        'drop_unused_frames': True,
    }

    out_paths = {
        'pred_pxl_fpath' : 'pred.kwcoco.zip',
    }

    @profile
    def command(self):
        fmtkw = self.final_config.copy()
        perf_config = self.final_perf_config
        algo_config = self.final_algo_config - {
            'package_fpath', 'test_dataset', 'pred_dataset'}
        fmtkw['params_argstr'] = self._make_argstr(algo_config)
        fmtkw['perf_argstr'] = self._make_argstr(perf_config)
        command = ub.codeblock(
            r'''
            python -m watch.tasks.fusion.predict \
                --package_fpath={package_fpath} \
                --test_dataset={test_dataset} \
                --pred_dataset={pred_pxl_fpath} \
                {params_argstr} \
                {perf_argstr}
            ''').format(**fmtkw).rstrip().rstrip('\\').rstrip()
        return command


class PolygonPrediction(ProcessNode):
    executable = 'python -m watch.cli.run_tracker'
    group_dname = PREDICT_NAME
    default_track_fn = NotImplemented

    in_paths = {
        'pred_pxl_fpath',
        'site_summary',
        'boundary_region',
    }

    # also
    # algo_params = {
    #     'resolution': None,
    # }

    out_paths = {
        'site_summaries_fpath': 'site_summaries_manifest.json',
        'site_summaries_dpath': 'site_summaries',
        'sites_fpath': 'sites_manifest.json',
        'sites_dpath': 'sites',
        'poly_kwcoco_fpath': 'poly.kwcoco.zip'
    }

    @profile
    def command(self):
        fmtkw = self.final_config.copy()
        fmtkw['default_track_fn'] = self.default_track_fn
        external_args = {'site_summary', 'boundary_region'}
        track_kwargs = self.final_algo_config.copy() - external_args
        fmtkw['kwargs_str'] = shlex.quote(json.dumps(track_kwargs))
        command = ub.codeblock(
            r'''
            python -m watch.cli.run_tracker \
                --in_file "{pred_pxl_fpath}" \
                --default_track_fn {default_track_fn} \
                --track_kwargs {kwargs_str} \
                --clear_annots=True \
                --site_summary '{site_summary}' \
                --boundary_region '{boundary_region}' \
                --out_site_summaries_fpath "{site_summaries_fpath}" \
                --out_site_summaries_dir "{site_summaries_dpath}" \
                --out_sites_fpath "{sites_fpath}" \
                --out_sites_dir "{sites_dpath}" \
                --out_kwcoco "{poly_kwcoco_fpath}"
            ''').format(**fmtkw)
        return command


class PolygonEvaluation(ProcessNode):
    executable = 'python -m watch.cli.run_metrics_framework'
    group_dname = EVALUATE_NAME

    in_paths = {
        'true_region_dpath',
        'true_site_dpath',
        'sites_fpath',
    }

    out_paths = {
        'eval_dpath': '.',
        'eval_fpath': 'poly_eval.json',
    }

    @profile
    def command(self):
        # self.tmp_dpath = self.paths['eval_dpath'] / 'tmp'
        # self.tmp_dpath = self.paths['eval_dpath'] / 'tmp'
        fmtkw = self.final_config.copy()

        handled = {
            'name', 'true_site_dpath', 'merge', 'true_region_dpath',
            'true_region_dpath', 'pred_sites', 'tmp_dir', 'out_dir',
            'merge_fpath',
        }

        fmtkw['params_argstr'] = self._make_argstr(self.final_algo_config - handled)
        fmtkw['perf_argstr'] = self._make_argstr(self.final_perf_config - handled)

        fmtkw['tmp_dpath'] = self.final_node_dpath / 'tmp'

        # Hack:
        if fmtkw['true_site_dpath'] is None:
            raise Exception('You must specify true_site_dpath and true_region_dpath')
            dvc_dpath = _phase2_dvc_data_dpath()
            fmtkw['true_site_dpath'] = dvc_dpath / 'annotations/site_models'
            fmtkw['true_region_dpath'] = dvc_dpath / 'annotations/region_models'

        name_parts = {
            k: v for k, v in sorted(self.condensed.items())
            if 'eval' not in k and (('algo_id' in k) or ('id' not in v))
        }
        fmtkw['name_suffix'] = '-'.join(name_parts.values())

        command = ub.codeblock(
            r'''
            python -m watch.cli.run_metrics_framework \
                --merge=True \
                --name "{name_suffix}" \
                --true_site_dpath "{true_site_dpath}" \
                --true_region_dpath "{true_region_dpath}" \
                --pred_sites "{sites_fpath}" \
                --tmp_dir "{tmp_dpath}" \
                --out_dir "{eval_dpath}" \
                --merge_fpath "{eval_fpath}" \
                {params_argstr} \
                {perf_argstr}
            ''').format(**fmtkw)
        command = command.rstrip().rstrip('\\').rstrip()
        return command


@ub.memoize
def _phase2_dvc_data_dpath():
    import watch
    dvc_dpath = watch.find_dvc_dpath(tags='phase2_data', hardware='auto')
    return dvc_dpath


class HeatmapEvaluation(ProcessNode):
    executable = 'python -m watch.tasks.fusion.evaluate'
    group_dname = EVALUATE_NAME

    in_paths = {
        'true_dataset',
        'pred_pxl_fpath',
    }

    out_paths = {
        'eval_pxl_dpath': '.',
        'eval_pxl_fpath': 'pxl_eval.json',
    }

    @profile
    def command(self):
        # TODO: better score space
        fmtkw = self.final_config.copy()
        extra_opts = {
            'draw_curves': True,
            'draw_heatmaps': False,
            'viz_thresh': 'auto',
            'workers': 2,
            'score_space': 'video',
        }
        fmtkw['extra_argstr'] = self._make_argstr(extra_opts)  # NOQA
        command = ub.codeblock(
            r'''
            python -m watch.tasks.fusion.evaluate \
                --true_dataset={true_dataset} \
                --pred_dataset={pred_pxl_fpath} \
                --eval_dpath={eval_pxl_dpath} \
                --eval_fpath={eval_pxl_fpath} \
                {extra_argstr}
            ''').format(**fmtkw)
        # .format(**eval_act_pxl_kw).strip().rstrip('\\')
        return command


class KWCocoVisualization(ProcessNode):
    executable = 'python -m watch.cli.coco_visualize_videos'
    group_dname = PREDICT_NAME

    resources = {
        'cpus': 2,
    }

    in_paths = {
        'poly_kwcoco_fpath',
    }

    out_paths = {
        'viz_dpath': '.',
        'viz_stamp_fpath': '_viz.stamp'
    }

    @profile
    def command(self):
        fmtkw = self.final_config.copy()
        name_parts = {
            k: v for k, v in sorted(self.condensed.items())
            if 'eval' not in k and (('algo_id' in k) or ('id' not in v))
        }
        fmtkw['name_suffix'] = '-'.join(name_parts.values())
        # paths = ub.udict(paths)
        # viz_pred_trk_poly_kw = paths.copy()
        # fmtkw['extra_header'] = f"\\n{condensed['trk_pxl_algo_id']}-{condensed['trk_poly_algo_id']}"
        # viz_pred_trk_poly_kw['viz_channels'] = "red|green|blue,salient"
        command = ub.codeblock(
            r'''
            smartwatch visualize \
                "{poly_kwcoco_fpath}" \
                --viz_dpath={viz_dpath} \
                --channels="auto" \
                --stack=only \
                --workers=2 \
                --extra_header="{name_suffix}" \
                --animate=True && touch {viz_stamp_fpath}
            ''').format(**fmtkw)
        return command


###
# Team Feature Nodes
###


class InvariantFeatureComputation(FeatureComputation):
    name = 'invar_feat'

    out_paths = {
        'dst': 'feat_I_{invar_feat_id}.kwcoco.zip'
    }


class MaterialFeatureComputation(FeatureComputation):
    name = 'mat_feat'

    out_paths = {
        'dst': 'feat_M_{mat_feat_id}.kwcoco.zip'
    }


class LandcoverFeatureComputation(FeatureComputation):
    name = 'land_feat'

    out_paths = {
        'dst': 'feat_L_{land_feat_id}.kwcoco.zip'
    }


###
# BAS / SC Nodes
###

# ---

class BAS_HeatmapPrediction(HeatmapPrediction):
    """

    CommandLine:
        xdoctest -m watch.mlops.smart_pipeline BAS_HeatmapPrediction

    Example:
        >>> from watch.mlops.smart_pipeline import *  # NOQA
        >>> node = BAS_HeatmapPrediction()
        >>> node.configure({
        >>>     'tta_time': 2,
        >>>     'package_fpath': 'foo.pt',
        >>>     'test_dataset': 'bar.json',
        >>> })
        >>> command = node.command()
        >>> assert 'tta_time="2"' in command
        >>> print(command)
    """
    name = 'bas_pxl'
    # # node_dname = 'bas_pxl/{bas_model}/{bas_test_dset}/{bas_pxl_algo_id}/{bas_pxl_id}'
    # node_dname = 'bas_pxl/{bas_pxl_algo_id}/{bas_pxl_id}'

    perf_params = ub.udict(HeatmapPrediction.perf_params) | {
        'with_saliency': True,
        'with_class': False,
        'with_change': False,
    }

    @property
    def condensed(self):
        condensed = super().condensed
        condensed['bas_model'] = 'todo'
        condensed['bas_test_dset'] = 'todo'
        return condensed


class SC_HeatmapPrediction(HeatmapPrediction):
    name = 'sc_pxl'
    # # node_dname = 'sc_pxl/{sc_model}/{sc_test_dset}/{sc_pxl_algo_id}/{sc_pxl_id}'
    # node_dname = 'sc_pxl/{sc_pxl_algo_id}/{sc_pxl_id}'

    perf_params = ub.udict(HeatmapPrediction.perf_params) | {
        'with_saliency': False,
        'with_class': True,
        'with_change': False,
    }

    @property
    def condensed(self):
        condensed = super().condensed
        condensed['sc_model'] = 'todo'
        condensed['sc_test_dset'] = 'todo'
        return condensed

# ---


class BAS_PolygonPrediction(PolygonPrediction):
    name = 'bas_poly'
    # node_dname = 'bas_poly/{bas_poly_algo_id}/{bas_poly_id}'
    default_track_fn = 'saliency_heatmaps'

    @property
    def final_algo_config(self):
        return ub.udict({
            # 'boundaries_as': 'polys'
            "agg_fn": "probs",
        }) | super().final_algo_config


class SC_PolygonPrediction(PolygonPrediction):
    name = 'sc_poly'
    # node_dname = 'sc_poly/{sc_poly_algo_id}/{sc_poly_id}'
    default_track_fn = 'class_heatmaps'

    @property
    def final_algo_config(self):
        return ub.udict({
            'boundaries_as': 'polys'
        }) | super().final_algo_config

# ---


class BAS_HeatmapEvaluation(HeatmapEvaluation):
    name = 'bas_pxl_eval'
    # node_dname = 'bas_pxl_eval'


class SC_HeatmapEvaluation(HeatmapEvaluation):
    name = 'sc_pxl_eval'
    # node_dname = 'sc_pxl_eval'


# ---

class BAS_PolygonEvaluation(PolygonEvaluation):
    name = 'bas_poly_eval'
    # node_dname = 'bas_poly_eval'


class SC_PolygonEvaluation(PolygonEvaluation):
    name = 'sc_poly_eval'
    # node_dname = 'sc_poly_eval'

# ---


class BAS_Visualization(KWCocoVisualization):
    name = 'bas_poly_viz'
    # node_dname = 'bas_poly_viz'


class SC_Visualization(KWCocoVisualization):
    name = 'sc_poly_viz'
    # node_dname = 'sc_poly_viz'


# ---


class Cropping(ProcessNode):
    """
    Used for both site cropping and validation-cropping
    """
    executable = 'python -m watch.cli.align'
    group_dname = PREDICT_NAME

    algo_params = {
        # 'include_channels': 'red|green|blue|cloudmask',  # fixme: not a good default
        'include_channels': None,
        'exclude_sensors': 'L8',  # fixme: not a good default
        'target_gsd': 4,
        # 'context_factor': 2.0,
        'context_factor': 1.5,
        'force_nodata': -9999,
        'rpc_align_method': 'orthorectify',
    }

    # The best setting of this depends on if the data is remote or not.  When
    # networking, around 20+ workers is a good idea, but that's a very bad idea
    # for local images or if the images are too big.
    perf_params = {
        'verbose': 1,
        # 'workers': 8,
        # 'aux_workers': 16,
        'workers': 32,
        'aux_workers': 4,
        'debug_valid_regions': False,
        'visualize': False,
        'keep': 'img',
        'geo_preprop': 'auto',
    }

    in_paths = {
        'crop_src_fpath',
        'regions',
    }

    @property
    def condensed(self):
        condensed = super().condensed
        condensed['regions_id'] = 'todo'
        condensed['src_dset'] = 'todo'
        return condensed

    @profile
    def command(self):
        fmtkw = self.final_config.copy()
        algo_config = self.final_algo_config - {'crop_src_fpath'}
        fmtkw['crop_algo_argstr'] = self._make_argstr(algo_config)
        fmtkw['crop_perf_argstr'] = self._make_argstr(self.final_perf_config)
        fmtkw.update(self.final_in_paths)
        fmtkw.update(self.final_out_paths)

        command = ub.codeblock(
            r'''
            python -m watch.cli.coco_align \
                --src "{crop_src_fpath}" \
                --dst "{crop_dst_fpath}" \
                --regions="{regions}" \
                --site_summary=True \
                {crop_perf_argstr} \
                {crop_algo_argstr}
            ''').format(**fmtkw)

        # FIXME: parametarize and only if we need secrets
        # secret_fpath = ub.Path('$HOME/code/watch/secrets/secrets').expand()
        # # if ub.Path.home().name.startswith('jon'):
        #     # if secret_fpath.exists():
        #     #     secret_fpath
        #         # command = f'source {secret_fpath} && ' + command
        # command = 'AWS_DEFAULT_PROFILE=iarpa GDAL_DISABLE_READDIR_ON_OPEN=EMPTY_DIR ' + command
        return command


class SC_Cropping(Cropping):
    """
    Crop to each image of every site.

    Example:
        >>> node = SC_Cropping()
        >>> print(node.command())
    """
    name = 'sc_crop'
    group_dname = 'crops'
    # node_dname = 'sitecrop/{src_dset}/{regions_id}/{sitecrop_algo_id}/{sitecrop_id}'

    algo_params = {
        # 'include_channels': 'red|green|blue|cloudmask',  # fixme: not a good default
        'include_channels': None,
        'exclude_sensors': 'L8',  # fixme: not a good default
        'target_gsd': 4,
        # 'context_factor': 2.0,
        'context_factor': 1.0,
        'force_nodata': -9999,
        'rpc_align_method': 'orthorectify',
        'convexify_regions': True,
        'minimum_size': '128x128@10GSD',
        'force_min_gsd': 2,
    }

    out_paths = {
        'crop_dst_fpath': 'sitecrop.kwcoco.zip'
    }


class SV_Cropping(Cropping):
    """
    Crop to high res images as the start / end of a sequence

    Example:
        >>> node = SV_Cropping()
        >>> print(node.command())
    """
    name = 'sv_crop'
    # node_dname = 'sv_crop/{src_dset}/{regions_id}/{valicrop_algoid}/{sitecrop_id}'

    algo_params = {
        # 'include_channels': 'red|green|blue|cloudmask',  # fixme: not a good default
        'include_sensors': 'WV',
        # None,
        # 'exclude_sensors': 'L8',  # fixme: not a good default
        # 'context_factor': 2.0,
        'context_factor': 1.3,
        'force_nodata': -9999,
        'rpc_align_method': 'orthorectify',
        'minimum_size': '128x128@2GSD',
        'target_gsd': 2,
        'force_min_gsd': 2,
        'convexify_regions': True,
        'num_end_frames': 3,
        'num_start_frames': 3,
    }

    out_paths = {
        'crop_dst_fpath': 'sv_crop.kwcoco.zip'
    }


class DinoBoxDetector(ProcessNode):
    """
    Used for both site cropping and validation-cropping

    Example:
        >>> node = DinoBoxDetector(root_dpath='/root/dpath/')
        >>> node.configure({
        >>>     'coco_fpath': 'foo.kwcoco',
        >>>     'package_fpath': 'model.pt',
        >>> })
        >>> print(node.command())
        >>> algo_id1 = node.algo_id
        >>> print(f'node.algo_id={node.algo_id}')
        >>> print(f'node.process_id={node.process_id}')
        >>> node.configure({
        >>>     'data_workers': 10,
        >>> })
        >>> algo_id2 = node.algo_id
        >>> node.configure({
        >>>     'fixed_resolution': "10GSD",
        >>> })
        >>> algo_id3 = node.algo_id
        >>> assert algo_id1 == algo_id2, 'perf params dont change hash'
        >>> assert algo_id1 != algo_id3, 'algo params do change hash'
    """
    name = 'sv_dino_boxes'
    executable = 'python -m watch.tasks.dino_detector.predict'
    group_dname = PREDICT_NAME

    in_paths = {
        'coco_fpath',
        'package_fpath',
    }
    out_paths = {
        'out_coco_fpath': 'pred_boxes.kwcoco.zip'
    }

    algo_params = {
        'fixed_resolution': "2GSD",
        'batch_size': 1,
    }

    # The best setting of this depends on if the data is remote or not.  When
    # networking, around 20+ workers is a good idea, but that's a very bad idea
    # for local images or if the images are too big.
    perf_params = {
        'device': 0,
        'data_workers': 2,
    }

    # @property
    # def condensed(self):
    #     condensed = super().condensed
    #     condensed['regions_id'] = 'todo'
    #     condensed['src_dset'] = 'todo'
    #     return condensed

    @profile
    def command(self):
        fmtkw = {}
        # Not sure why final-config doesn't have everything
        config = (ub.udict(self.final_config) | self.final_algo_config) | self.final_perf_config
        assert config['package_fpath']  is not None
        fmtkw['config_argstr'] = self._make_argstr(config)
        command = ub.codeblock(
            r'''
            python -m watch.tasks.dino_detector.predict \
                {config_argstr}
            ''').format(**fmtkw)
        return command


# from watch.tasks.dino_detector import building_validator  # NOQA
# ub.udict(building_validator.BuildingValidatorConfig.__default__)

class DinoBuildingFilter(ProcessNode):
    """
    Used for both site cropping and validation-cropping

    Example:
        >>> from watch.mlops.smart_pipeline import *  # NOQA
        >>> self = node = DinoBuildingFilter(root_dpath='/ROOT/DPATH/')
        >>> node.configure({
        >>>     'input_kwcoco': 'foo.kwcoco',
        >>>     'input_region': 'region.geojson',
        >>>     'input_sites': 'input_sites',
        >>>     #'output_sites_dpath': 'I_WANT_OUT_SITES_HERE',
        >>>     'output_region_fpath': 'I_WANT_OUT_REGIONS_HERE',
        >>>     'output_site_manifest_fpath': 'I_WANT_SITE_MANIFESTS_HERE',
        >>> })
        >>> print('self.template_out_paths = {}'.format(ub.urepr(self.template_out_paths, nl=1)))
        >>> print('self.final_out_paths = {}'.format(ub.urepr(self.final_out_paths, nl=1)))
        >>> print(node.command())
    """
    name = 'sv_dino_filter'
    executable = 'python -m watch.tasks.dino_detector.building_validator'
    group_dname = PREDICT_NAME

    in_paths = {
        'input_kwcoco',
        'input_region',
        'input_sites',
    }
    out_paths = {
        'output_region_fpath': 'out_region.geojson',
        'output_sites_dpath': 'out_sites',
        'output_site_manifest_fpath': 'out_site_manifest.json',
    }

    algo_params = {
        'box_isect_threshold': 0.1,
        'box_score_threshold': 0.1,
        'start_max_score': 1.0,
        'end_min_score': 0.1,
    }

    @profile
    def command(self):
        # Not sure why final-config doesn't have everything
        config = (ub.udict(self.final_config) | self.final_algo_config) | self.final_perf_config
        config_argstr = self._make_argstr(config)
        command = ub.codeblock(
            r'''
            {executable} \
                {config_argstr}
            ''').format(
                executable=self.executable,
                config_argstr=config_argstr,
            )
        return command


def bas_nodes():
    nodes = {}

    nodes['bas_pxl'] = BAS_HeatmapPrediction()
    nodes['bas_poly'] = BAS_PolygonPrediction()
    nodes['bas_pxl_eval'] = BAS_HeatmapEvaluation()
    nodes['bas_poly_eval'] = BAS_PolygonEvaluation()
    nodes['bas_poly_viz'] = BAS_Visualization()

    nodes['bas_pxl'].inputs['test_dataset'].connect(
        nodes['bas_pxl_eval'].inputs['true_dataset']
    )

    nodes['bas_pxl'].connect(
        nodes['bas_pxl_eval'],
        nodes['bas_poly'],
    )
    nodes['bas_poly'].connect(
        nodes['bas_poly_eval'],
        nodes['bas_poly_viz'],
    )

    if 0:
        nodes['bas_invariants'] = InvariantFeatureComputation()
        nodes['bas_land'] = LandcoverFeatureComputation()
        nodes['bas_materials'] = MaterialFeatureComputation()
        nodes['bas_featunion'] = FeatureUnion()

        nodes['bas_invariants'].connect(
            nodes['bas_featunion'],
            src_map={'dst': '_'},
            dst_map={'src': '_'},
        )
        nodes['bas_land'].connect(
            nodes['bas_featunion'],
            src_map={'dst': '_'},
            dst_map={'src': '_'},
        )
        nodes['bas_materials'].connect(
            nodes['bas_featunion'],
            src_map={'dst': '_'},
            dst_map={'src': '_'},
        )

        nodes['bas_featunion'].connect(nodes['bas_pxl'], param_mapping={
            'dst': 'test_dataset'
        })

        # nodes['bas_featunion'].connect(nodes['bas_pxl'], param_mapping={
        #     ''
        # })
        # nodes['bas_land'].connect(nodes['bas_pxl'])
        # nodes['bas_materials'].connect(nodes['bas_pxl'])
    return nodes


def sc_nodes():
    nodes = {}
    nodes['sc_pxl'] = SC_HeatmapPrediction()
    nodes['sc_poly'] = SC_PolygonPrediction()
    nodes['sc_pxl_eval'] = SC_HeatmapEvaluation()
    nodes['sc_poly_eval'] = SC_PolygonEvaluation()
    nodes['sc_poly_viz'] = SC_Visualization()

    nodes['sc_pxl'].inputs['test_dataset'].connect(
        nodes['sc_pxl_eval'].inputs['true_dataset']
    )

    nodes['sc_pxl'].connect(
        nodes['sc_pxl_eval'],
        nodes['sc_poly'].connect(
            nodes['sc_poly_eval'],
            nodes['sc_poly_viz'],
        ),
    )
    return nodes


def make_smart_pipeline_nodes(with_bas=True, building_validation=False,
                              site_crops=True, with_sc=True):
    nodes = {}

    sc_input_region = None
    sc_input_kwcoco = None
    bas_input_kwcoco = None

    true_regions = None
    true_sites = None

    if with_bas:
        nodes.update(bas_nodes())
        bas_output_region = nodes['bas_poly'].outputs['site_summaries_fpath']
        bas_output_sites = nodes['bas_poly'].outputs['sites_fpath']

        bas_input_kwcoco = nodes['bas_pxl'].inputs['test_dataset']
        sc_input_kwcoco = bas_input_kwcoco

        # By default the input regions to SC are BAS outputs
        sc_input_region = bas_output_region

        # By default SC will use the same input as BAS
        sc_input_kwcoco = sc_input_kwcoco

        true_regions = nodes['bas_poly_eval'].inputs['true_region_dpath']
        true_sites = nodes['bas_poly_eval'].inputs['true_site_dpath']

    if building_validation:
        nodes['sv_crop'] = SV_Cropping()

        if bas_input_kwcoco is not None:
            bas_output_region.connect(nodes['sv_crop'].inputs['regions'])
            bas_input_kwcoco.connect(nodes['sv_crop'].inputs['crop_src_fpath'])

        nodes['sv_dino_boxes'] = DinoBoxDetector()
        nodes['sv_crop'].outputs['crop_dst_fpath'].connect(nodes['sv_dino_boxes'].inputs['coco_fpath'])

        nodes['sv_dino_filter'] = DinoBuildingFilter()
        nodes['sv_dino_boxes'].outputs['out_coco_fpath'].connect(nodes['sv_dino_filter'].inputs['input_kwcoco'])
        bas_output_region.connect(nodes['sv_dino_filter'].inputs['input_region'])
        bas_output_sites.connect(nodes['sv_dino_filter'].inputs['input_sites'])

        # If we have a validation step, use those as inputs to SC instead
        sv_output_region = nodes['sv_dino_filter'].outputs['output_region_fpath']
        sv_output_sites = nodes['sv_dino_filter'].outputs['output_sites_dpath']

        # Add an evaluation step after bas validation
        nodes['sv_poly_eval'] = PolygonEvaluation(name='sv_poly_eval')
        sv_output_sites.connect(nodes['sv_poly_eval'].inputs['sites_fpath'])
        sc_input_region = sv_output_region

        if true_regions is not None:
            true_regions.connect(nodes['sv_poly_eval'].inputs['true_region_dpath'])
            true_sites.connect(nodes['sv_poly_eval'].inputs['true_site_dpath'])

    if site_crops:
        nodes['sc_crop'] = SC_Cropping()
        # If we predicted regions to go to SC, crop to those.
        if sc_input_region is not None:
            sc_input_region.connect(nodes['sc_crop'].inputs['regions'])
            # nodes['bas_pxl'].inputs['test_dataset'].connect(sc_input_kwcoco)
            # If we are site cropping, then use its outputs as SC input
            sc_input_kwcoco = nodes['sc_crop'].outputs['crop_dst_fpath']

    if with_sc:
        nodes.update(sc_nodes())

        if sc_input_kwcoco is not None:
            sc_input_kwcoco.connect(nodes['sc_pxl'].inputs['test_dataset'])

        if sc_input_region is not None:
            sc_input_region.connect(nodes['sc_poly'].inputs['site_summary'])

        if true_regions is not None:
            true_regions.connect(nodes['sc_poly_eval'].inputs['true_region_dpath'])
            true_sites.connect(nodes['sc_poly_eval'].inputs['true_site_dpath'])
    return nodes


def make_smart_pipeline(name):
    """
    Get an unconfigured instance of the SMART pipeline

    CommandLine:
        xdoctest -m watch.mlops.smart_pipeline make_smart_pipeline

    Example:
        >>> from watch.mlops.smart_pipeline import *  # NOQA
        >>> dag = make_smart_pipeline('bas_building_vali')
        >>> dag.print_graphs()
    """
    from watch.mlops.pipeline_nodes import PipelineDAG
    from functools import partial
    node_makers = {
        'joint_bas_sc': partial(make_smart_pipeline_nodes, site_crops=True),
        'joint_bas_sc_nocrop': partial(make_smart_pipeline_nodes, site_crops=False),
        'crop_sc': partial(make_smart_pipeline_nodes, with_bas=False, site_crops=True),
        'sc': sc_nodes,
        'bas': bas_nodes,
        'bas_building_vali': partial(make_smart_pipeline_nodes, with_bas=True,
                                     building_validation=True,
                                     site_crops=False, with_sc=False),
    }
    make_nodes = node_makers[name]
    nodes = make_nodes()
    dag = PipelineDAG(nodes)
    dag.build_nx_graphs()
    return dag


# from xdev import profile  # NOQA
# profile.add_module()



#!/usr/bin/env python3
"""
See Old Version:
    ~/code/watch/scripts/run_copy_previous_region_model_pseudolive.py

SeeAlso:
    ~/code/watch-smartflow-dags/KIT_TA2_PREEVAL10_PYENV_V13.py
"""
import os
import subprocess
import json
import ubelt as ub
import scriptconfig as scfg

from watch.utils.util_framework import download_region


class PsuedoliveCopyPreviousConfig(scfg.DataConfig):
    """
    Download, update, and upload region model for pseudo-live
    """
    input_region_path = scfg.Value(None, type=str, position=1, required=True, help=ub.paragraph(
            '''
            Path to input T&E Baseline Framework Region definition JSON
            '''))
    aws_profile = scfg.Value(None, type=str, help=ub.paragraph(
            '''
            AWS Profile to use for AWS S3 CLI commands
            '''))
    outbucket = scfg.Value(None, type=str, required=True, short_alias=['o'], help=ub.paragraph(
            '''
            S3 Output directory for STAC item / asset egress
            '''))
    updated_end_date = scfg.Value(None, type=str, required=True, short_alias=['e'], help=ub.paragraph(
            '''
            Set uploaded region model end date to this value
            '''))


def main():
    config = PsuedoliveCopyPreviousConfig.cli(strict=True)
    print('config = {}'.format(ub.urepr(dict(config), nl=1, align=':')))
    run_copy_previous_region_model(**config)


def _update_region(region, updated_end_date):
    for feature in region.get('features', ()):
        if feature['properties']['type'] == 'region':
            feature['properties']['end_date'] = updated_end_date
            break

    return region


def run_copy_previous_region_model(input_region_path,
                                   outbucket,
                                   updated_end_date,
                                   aws_profile=None):
    if aws_profile is not None:
        aws_base_command = ['aws', 's3', '--profile', aws_profile, 'cp']
    else:
        aws_base_command = ['aws', 's3', 'cp']

    # 1. Download region file
    print("* Downloading and pruning region file *")
    local_region_path = '/tmp/region.json'
    local_region_path = download_region(input_region_path,
                                        local_region_path,
                                        aws_profile=aws_profile,
                                        strip_nonregions=False)

    with open(local_region_path) as f:
        region = json.load(f)

    updated_region = _update_region(region, updated_end_date)

    local_updated_region_path = '/tmp/updated_region.json'
    with open(local_updated_region_path, 'w') as f:
        print(json.dumps(updated_region, indent=2), file=f)

    # 2. Upload updated region file
    print("* Uploading updated region file *")
    upload_region_path = os.path.join(
        outbucket, os.path.basename(input_region_path))
    subprocess.run([*aws_base_command,
                    local_updated_region_path,
                    upload_region_path], check=True)


if __name__ == "__main__":
    main()



import os
from datetime import datetime
from collections import namedtuple

import torch
import kwcoco
import kwimage
import numpy as np
from matplotlib.colors import to_rgba

from watch.tasks.rutgers_material_change_detection.utils.util_misc import get_crop_slices
from watch.tasks.rutgers_material_change_detection.datasets.base_dataset import BaseDataset


class IARPA_KWDATASET(BaseDataset):
    def __init__(
        self,
        root_dir,
        split,
        video_slice,
        task_mode,
        seed_num=0,
        sensor_type="S2",
        channels=None,
        transforms=None,
        normalize_mode=None,
        max_iterations=None,
    ):
        """Constructor.

        Args:
            kwcoco_path (str): File path to kwcoco file.
            video_slice (namedtuple):
            task_mode (str, str): Name of labels to produce for this dataset. Defaults to 'total_bin_change'.
            sensor_type (str, optional): The name of sensor to return image data from. Defaults to 'S2'.
        """
        super().__init__(
            root_dir,
            split,
            video_slice,
            task_mode,
            transforms=transforms,
            seed_num=seed_num,
            normalize_mode=normalize_mode,
            channels=channels,
            max_iterations=max_iterations,
        )

        # Figure out how many channels to load.
        Channel_Info = namedtuple("channel_info", ["band_name", "wl_name", "wavelength", "scale_factor"])
        self.channel_info = {}
        if channels == "RGB":
            self.channel_info["B02"] = Channel_Info(band_name="B02", wl_name="blue", wavelength=492.4, scale_factor=1)
            self.channel_info["B03"] = Channel_Info(band_name="B03", wl_name="green", wavelength=559.8, scale_factor=1)
            self.channel_info["B04"] = Channel_Info(band_name="B04", wl_name="red", wavelength=664.6, scale_factor=1)
        elif channels == "RGB_NIR":
            self.channel_info["B02"] = Channel_Info(band_name="B02", wl_name="blue", wavelength=492.4, scale_factor=1)
            self.channel_info["B03"] = Channel_Info(band_name="B03", wl_name="green", wavelength=559.8, scale_factor=1)
            self.channel_info["B04"] = Channel_Info(band_name="B04", wl_name="red", wavelength=664.6, scale_factor=1)
            self.channel_info["B08"] = Channel_Info(band_name="B08", wl_name="nir", wavelength=832.8, scale_factor=1)
        elif channels == "ALL":
            self.channel_info["B01"] = Channel_Info(band_name="B01", wl_name="coastal", wavelength=442, scale_factor=6)
            self.channel_info["B02"] = Channel_Info(band_name="B02", wl_name="blue", wavelength=492, scale_factor=1)
            self.channel_info["B03"] = Channel_Info(band_name="B03", wl_name="green", wavelength=559, scale_factor=1)
            self.channel_info["B04"] = Channel_Info(band_name="B04", wl_name="red", wavelength=664, scale_factor=1)
            self.channel_info["B05"] = Channel_Info(band_name="B05", wl_name="nir", wavelength=704, scale_factor=2)
            self.channel_info["B06"] = Channel_Info(band_name="B06", wl_name="nir", wavelength=740, scale_factor=2)
            self.channel_info["B07"] = Channel_Info(band_name="B07", wl_name="nir", wavelength=783, scale_factor=2)
            self.channel_info["B08"] = Channel_Info(band_name="B08", wl_name="nir", wavelength=832, scale_factor=1)
            self.channel_info["B09"] = Channel_Info(band_name="B09", wl_name="swir", wavelength=945, scale_factor=6)
            self.channel_info["B10"] = Channel_Info(band_name="B10", wl_name="swir", wavelength=1375, scale_factor=6)
            self.channel_info["B11"] = Channel_Info(band_name="B11", wl_name="swir", wavelength=1610, scale_factor=2)
            self.channel_info["B12"] = Channel_Info(band_name="B12", wl_name="swir", wavelength=2190, scale_factor=2)
            self.channel_info["B8A"] = Channel_Info(band_name="B8A", wl_name="nir", wavelength=864, scale_factor=2)
        else:
            raise NotImplementedError(f'Channels equal to "{channels}" not implmented.')
        self.n_channels = len(self.channel_info.keys())

        # Get number of classes for dataset with given task mode.
        if task_mode == "total_bin_change":
            self.n_classes = 2
        elif task_mode == "ss_mat_recon":
            self.n_classes = self.n_channels
        elif task_mode == "total_sem_change":
            self.n_classes = 15
        else:
            raise NotImplementedError(f'Number of classes for this task "{task_mode}" needs to be added.')

        if split == "train":
            kwcoco_path = os.path.join(root_dir, "train_data.kwcoco.json")
        elif split == "valid":
            kwcoco_path = os.path.join(root_dir, "vali_data.kwcoco.json")
        else:
            raise NotImplementedError(f'Mode "{split}" not implemented for IARPA KWCOCO Dataset class.')

        self.dset_path = os.path.split(kwcoco_path)[0]
        self.transforms = transforms
        self.task_mode = task_mode
        self._sensor_type = sensor_type

        # Load kwcoco dataset
        self.coco_dset = kwcoco.CocoDataset(kwcoco_path)

        # Get all video ids
        video_ids = [video_info["id"] for video_info in self.coco_dset.videos().objs]

        # Only get videos and images with desired sensor
        self.video_dataset = {}  # video_id -> image_ids (filtered)
        self.max_frames = 0
        for video_id in video_ids:
            # Get image ids for this video
            image_ids = self.coco_dset.index.vidid_to_gids[video_id]

            # Filter based on desired sensor type
            sesnor_image_ids = []
            for image_id in image_ids:
                img_sensor_type = self.coco_dset.index.imgs[image_id]["sensor_coarse"]
                if img_sensor_type == sensor_type:
                    sesnor_image_ids.append(image_id)

            # Update maximum number of frames.
            self.max_frames = max(len(sesnor_image_ids), self.max_frames)

            # Add to video dataset
            if len(sesnor_image_ids) > 1:
                self.video_dataset[video_id] = sesnor_image_ids

        # Create all slice versions
        self.examples = []
        for video_id, image_ids in self.video_dataset.items():
            # Get video shape
            H = self.coco_dset.index.imgs[image_ids[0]]["height"]
            W = self.coco_dset.index.imgs[image_ids[0]]["width"]
            crop_slices = get_crop_slices(
                H, W, self.video_slice.height, self.video_slice.width, step=self.video_slice.stride
            )

            # Get region ID.
            region_name = self.coco_dset.index.videos[video_id]["name"]

            # Poplate slices with more information.
            for crop_slice in crop_slices:
                crop_info = {
                    "video_id": video_id,
                    "og_height": H,
                    "og_width": W,
                    "space_crop_slice": crop_slice,
                    "region_name": region_name,
                }
                self.examples.append(crop_info)

        # Update max frames based on the number of frames.
        if self.n_frames is None:
            self.n_frames = self.max_frames

        if self.max_frames > self.n_frames:
            self.max_frames = self.n_frames

    def cid_to_name(self, cid):
        return self.coco_dset.cats[cid]["name"]

    def cid_to_rgb(self, cid):
        return to_rgba(self.coco_dset.cats[cid]["color"])[:3]

    def load_frames(self, image_ids, crop_slice):
        h0, w0, h, w = crop_slice

        # Get channels.
        if self.channels == "RGB":
            channel_names = "red|blue|green"
        elif self.channels == "RGB_NIR":
            channel_names = "red|blue|green|nir"
        elif self.channels == "ALL":
            channel_names = None

        # Load all of the images and slice according to crop
        image_crops = []
        for image_id in image_ids:
            delayed_crop = self.coco_dset.delayed_load(image_id, channels=channel_names, space="video").crop(
                (slice(h0, h0 + h), slice(w0, w0 + w))
            )
            delayed_crop = delayed_crop.finalize(as_xarray=False).transpose((2, 0, 1))
            image_crops.append(delayed_crop)

        return image_crops, image_ids

    def subset_image_ids(self, image_ids):

        # Sort image_ids by datetimes.
        datetimes = [self.get_datetime(image_id) for image_id in image_ids]
        image_ids = [img_id for dt, img_id in sorted(zip(datetimes, image_ids))]

        if len(image_ids) <= self.max_frames:
            # Grab all of the frames.
            new_image_ids = image_ids
        elif self.max_frames == 2:
            new_image_ids = [image_ids[0], image_ids[-1]]
        elif len(image_ids) > self.max_frames:
            if self.split == "train":
                if self.task_mode == "ss_mat_recon":
                    # Randomly sample some of the image ids from a video.
                    new_image_ids = list(np.random.choice(image_ids, self.max_frames))
                else:
                    # Randomly sample from indices between (0, last).
                    indices = np.arange(1, len(image_ids) - 1)
                    frame_indices = sorted(np.random.choice(indices, size=self.max_frames - 2, replace=False))
                    new_image_ids = list(np.take(np.asarray(image_ids), frame_indices))
                    new_image_ids.insert(0, image_ids[0])
                    new_image_ids.append(image_ids[-1])
            else:
                n_avail_img_ids = len(image_ids) - 2
                n_sub_frames = self.max_frames - 2
                indices = list(range(n_avail_img_ids))

                slice = n_avail_img_ids // n_sub_frames

                indices = indices[slice // 2 :: slice]
                indices.insert(0, 0)
                indices.append(-1)
                new_image_ids = np.take(image_ids, indices)

        return new_image_ids

    def get_total_binary_change_example(self, index):
        crop_info = self.examples[index]

        # Get all image ids for this video
        image_ids = self.video_dataset[crop_info["video_id"]]
        image_ids = self.subset_image_ids(image_ids)
        frames, selected_image_ids = self.load_frames(image_ids, crop_info["space_crop_slice"])

        ## Stack images
        video = np.stack(frames, axis=0)  # shape: [frames, channels, height, width]

        ## Format image values (remove nans, scale to 0-1)
        # Remove NaNs in images
        nan_indices = np.isnan(video)
        video[nan_indices] = 0

        # Scale from uint16 to [0,1].
        video = np.clip(video, 0, 2**15) / 2**15
        video = video.astype("float32")

        # Compute target label mask.
        target, ignore_mask = self.compute_target_mask(
            image_ids,
            self.task_mode,
            og_height=crop_info["og_height"],
            og_width=crop_info["og_width"],
            spatial_crop=crop_info["space_crop_slice"],
        )

        # Apply transforms to imagery.
        if self.transforms:
            # Need to create custom transform object
            video, target, ignore_mask = self.transforms(video, target, ignore_mask)

        # Normalize video.
        video = self.normalize(video)

        # Pad video to desired window size.
        if (video.shape[2] < self.video_slice.height) or (video.shape[3] < self.video_slice.width):
            canvas = (
                np.ones(
                    (video.shape[0], video.shape[1], self.video_slice.height, self.video_slice.width), dtype=video.dtype
                )
                * self.pad_value
            )
            canvas[:, :, : video.shape[2], : video.shape[3]] = video
            video = canvas

        # Pad target mask to desired window size.
        if (target.shape[0] < self.video_slice.height) or (target.shape[1] < self.video_slice.width):
            canvas = np.ones((self.video_slice.height, self.video_slice.width), dtype=target.dtype) * self.pad_value
            canvas[: target.shape[0], : target.shape[1]] = target
            target = canvas

        # Pad target mask to desired window size.
        if (ignore_mask.shape[0] < self.video_slice.height) or (ignore_mask.shape[1] < self.video_slice.width):
            canvas = (
                np.ones((self.video_slice.height, self.video_slice.width), dtype=ignore_mask.dtype) * self.pad_value
            )
            canvas[: ignore_mask.shape[0], : ignore_mask.shape[1]] = ignore_mask
            ignore_mask = canvas

        # Scale resolution of video and target.
        if self.scale != 1 and self.scale is not None:
            video, target = self.scale_video_target(video, target, self.scale)

        # Convert from numpy array to tensor.
        video, target, ignore_mask = np.copy(video), np.copy(target), np.copy(ignore_mask)
        video = torch.as_tensor(video)
        target = torch.as_tensor(target)
        ignore_mask = torch.as_tensor(ignore_mask)

        ## Pad the number of frames to maximum number of frames.
        # [frames, channels, height, width]
        active_frames = torch.zeros(self.max_frames)
        active_frames[: video.shape[0]] = torch.ones(video.shape[0])

        pad_video = (
            torch.ones([self.max_frames, video.shape[1], video.shape[2], video.shape[3]], dtype=video.dtype)
            * self.pad_value
        )
        pad_video[: video.shape[0], ...] = video
        video = pad_video

        ## Get metadata from all images in video.
        datetimes = []
        for image_id in image_ids:
            datetimes.append(self.get_datetime(image_id))

        output = {}
        output["video"] = video.float()
        output[self.task_mode] = target
        output["datetimes"] = datetimes
        output["active_frames"] = active_frames
        output["region_name"] = crop_info["region_name"]
        output["crop_info"] = crop_info
        output["ignore_mask"] = ignore_mask
        return output

    def get_total_semantic_change_example(self, index):
        crop_info = self.examples[index]

        # Get all image ids for this video
        image_ids = self.video_dataset[crop_info["video_id"]]
        image_ids = self.subset_image_ids(image_ids)
        frames, selected_image_ids = self.load_frames(image_ids, crop_info["space_crop_slice"])

        ## Stack images
        video = np.stack(frames, axis=0)  # shape: [frames, channels, height, width]

        ## Format image values (remove nans, scale to 0-1)
        # Remove NaNs in images
        nan_indices = np.isnan(video)
        video[nan_indices] = 0

        # Scale from uint16 to [0,1].
        video = np.clip(video, 0, 2**15) / 2**15

        # Compute target label mask.
        target = self.compute_target_mask(
            image_ids,
            self.task_mode,
            og_height=crop_info["og_height"],
            og_width=crop_info["og_width"],
            spatial_crop=crop_info["space_crop_slice"],
        )

        # Apply transforms to imagery.
        if self.transforms:
            # Need to create custom transform object
            video, target = self.transforms(video, target)

        # Normalize video.
        video = self.normalize(video)

        # Pad video to desired window size.
        if (video.shape[2] < self.video_slice.height) or (video.shape[3] < self.video_slice.width):
            canvas = (
                np.ones(
                    (video.shape[0], video.shape[1], self.video_slice.height, self.video_slice.width), dtype=video.dtype
                )
                * self.pad_value
            )
            canvas[:, :, : video.shape[2], : video.shape[3]] = video
            video = canvas

        # Pad target mask to desired window size.
        if (target.shape[0] < self.video_slice.height) or (target.shape[1] < self.video_slice.width):
            canvas = np.ones((self.video_slice.height, self.video_slice.width), dtype=target.dtype) * self.pad_value
            canvas[: target.shape[0], : target.shape[1]] = target
            target = canvas

        # Scale resolution of video and target.
        if self.scale != 1 and self.scale is not None:
            video, target = self.scale_video_target(video, target, self.scale)

        # Convert from numpy array to tensor.
        video, target = np.copy(video), np.copy(target)
        video = torch.as_tensor(video)
        target = torch.as_tensor(target)

        ## Pad the number of frames to maximum number of frames.
        # [frames, channels, height, width]
        active_frames = torch.zeros(self.max_frames)
        active_frames[: video.shape[0]] = torch.ones(video.shape[0])

        pad_video = (
            torch.ones([self.max_frames, video.shape[1], video.shape[2], video.shape[3]], dtype=video.dtype)
            * self.pad_value
        )
        pad_video[: video.shape[0], ...] = video
        video = pad_video

        ## Get metadata from all images in video.
        datetimes = []
        for image_id in image_ids:
            datetimes.append(self.get_datetime(image_id))

        output = {}
        output["video"] = video.float()
        output[self.task_mode] = target
        output["datetimes"] = datetimes
        output["active_frames"] = active_frames
        output["region_name"] = crop_info["region_name"]
        output["crop_info"] = crop_info
        return output

    def get_self_supervised_material_reconstruction_example(self, index):
        crop_info = self.examples[index]

        # Get all image ids for this video
        image_ids = self.video_dataset[crop_info["video_id"]]
        image_ids = self.subset_image_ids(image_ids)
        frames, selected_image_ids = self.load_frames(image_ids, crop_info["space_crop_slice"])

        ## Stack images
        video = np.stack(frames, axis=0)  # shape: [frames, channels, height, width]

        ## Format image values (remove nans, scale to 0-1)
        # Remove NaNs in images
        nan_indices = np.isnan(video)
        video[nan_indices] = 0

        # Scale from uint16 to [0,1].
        video = np.clip(video, 0, 2**15) / 2**15
        video = video.astype("float32")

        # Apply transforms to imagery.
        if self.transforms:
            # Need to create custom transform object
            video = self.transforms(video)

        # Normalize video.
        video = self.normalize(video)

        # Pad video to desired size.
        active_frames = torch.zeros(self.max_frames)
        active_frames[: video.shape[0]] = torch.ones(video.shape[0])

        if (video.shape[2] < self.video_slice.height) or (video.shape[3] < self.video_slice.width):
            canvas = (
                np.ones(
                    (video.shape[0], video.shape[1], self.video_slice.height, self.video_slice.width), dtype=video.dtype
                )
                * self.pad_value
            )
            canvas[:, :, : video.shape[2], : video.shape[3]] = video
            video = canvas

        pad_video = (
            np.ones([self.max_frames, video.shape[1], video.shape[2], video.shape[3]], dtype=video.dtype)
            * self.pad_value
        )
        pad_video[: video.shape[0], ...] = video
        video = pad_video

        # Scale resolution.
        if self.scale != 1 and self.scale is not None:
            video, _ = self.scale_video_target(video, None, self.scale)

        # Convert from numpy array to tensor.
        video = np.copy(video)
        video = torch.from_numpy(video)

        # Get metadata from all images in video.
        datetimes = []
        for image_id in image_ids:
            datetimes.append(self.get_datetime(image_id))

        # Create output dictionary.
        output = {}
        output["video"] = video
        output[self.task_mode] = video
        output["datetimes"] = datetimes
        output["active_frames"] = active_frames
        return output

    def get_datetime(self, image_id):
        date_captured = self.coco_dset.index.imgs[image_id]["date_captured"]  # '2017-03-23T09:30:31'
        try:
            datetime_object = datetime.strptime(date_captured[:-6], "%Y-%m-%dT%H:%M:%S")
        except ValueError:
            datetime_object = datetime.strptime(date_captured[:-6], "%Y-%m-%dT%H")
        return datetime_object

    def compute_target_mask(self, image_ids, mode, og_height, og_width, spatial_crop):
        h0, w0, h, w = spatial_crop

        # Sort image_ids by datetimes.
        datetimes = [self.get_datetime(image_id) for image_id in image_ids]
        image_ids = [img_id for dt, img_id in sorted(zip(datetimes, image_ids))]

        if mode == "total_bin_change":
            # Compute the change between first and last annotation image in video.

            # Get first and last annotation images.
            # first_anno, first_ignore = self.compute_segmentation_anno(image_ids[0], og_height, og_width)
            # TODO: Assume that the labels are propagated.
            final_anno, final_ignore = self.compute_segmentation_anno(image_ids[-1], og_height, og_width)

            # Crop annotations according to slice
            # first_anno = first_anno[h0:(h0 + h), w0:(w0 + w)]
            # first_ignore = first_ignore[h0:(h0 + h), w0:(w0 + w)]
            final_anno = final_anno[h0 : (h0 + h), w0 : (w0 + w)]
            final_ignore = final_ignore[h0 : (h0 + h), w0 : (w0 + w)]

            # Compute the difference between annotations.
            target_mask = np.zeros(final_anno.shape, dtype=final_anno.dtype)
            target_mask = np.clip(final_anno, 0, 1)
            # x, y = np.where(first_anno == final_anno)  # find where labels match (no-change)
            # target_mask[x, y] = 0

            # Set ignore labels.
            x, y = np.where(final_ignore == 1)
            # x, y = np.where((first_ignore == 1) & (final_ignore == 1))
            target_mask[x, y] = -1
            ignore_mask = final_ignore

        elif mode == "total_sem_change":
            # Get first and last annotation images.
            final_anno, final_ignore = self.compute_segmentation_anno(image_ids[-1], og_height, og_width)

            # Crop annotations according to slice
            final_anno = final_anno[h0 : (h0 + h), w0 : (w0 + w)]
            final_ignore = final_ignore[h0 : (h0 + h), w0 : (w0 + w)]

            # Set ignore labels.
            x, y = np.where(final_ignore == 1)
            final_anno[x, y] = -1

            target_mask = final_anno

        elif mode == "bas":
            breakpoint()
            pass

        elif mode == "pw_seq_binary":
            # Compute the change between each sequential pair of annotations
            assert NotImplementedError(mode)
        elif mode == "pw_all_binary":
            # Compute the change between all combinations of annotations
            # TODO: Figure out how to handle this properly
            assert NotImplementedError(mode)
        elif mode == "segmentation":
            # Return the propagated construction stage masks
            assert NotImplementedError(mode)
        else:
            assert NotImplementedError(f"Invalid mode type: {mode}")

        return target_mask, ignore_mask

    def compute_segmentation_anno(self, image_id, og_height, og_width):
        anno_ids = self.coco_dset.index.gid_to_aids[image_id]
        dets = kwimage.Detections.from_coco_annots(self.coco_dset.annots(anno_ids).objs, dset=self.coco_dset)

        warp_info = self.coco_dset.imgs[image_id]["warp_img_to_vid"]
        dets = dets.scale(warp_info["scale"])
        if "offset" in warp_info.keys():
            dets = dets.translate(warp_info["offset"])

        anno_canvas = np.zeros([og_height, og_width], dtype=np.int32)
        ignore_canvas = np.zeros([og_height, og_width], dtype=np.int32)
        ann_polys = dets.data["segmentations"].to_polygon_list()
        ann_cids = [dets.classes.idx_to_id[cidx] for cidx in dets.data["class_idxs"]]

        for polygon, cat_id in zip(ann_polys, ann_cids):
            cid_name = self.cid_to_name(cat_id).lower()
            if cid_name in ["ignore", "unknown", "negative"]:
                polygon.fill(ignore_canvas, value=1)
            else:
                polygon.fill(anno_canvas, value=cat_id)

        return anno_canvas, ignore_canvas

    def to_rgb(self, image, gamma=0.9):
        # Get RGB bands.
        if self.n_channels in [3, 4]:
            r_band = image[0] * 0.9
            g_band = image[1] * 1.0
            b_band = image[2] * 0.9
        elif self.n_channels == 13:
            r_band = image[3]
            g_band = image[2]
            b_band = image[1]
        else:
            raise NotImplementedError(f'Cannot handle "{self.n_channels}" number of channels in to_rgb method.')

        rgb_image = np.stack((r_band, g_band, b_band), axis=2)

        if "TA1" in self.dset_dir:
            rgb_image *= 8
        else:
            rgb_image *= 8

        # Adjust color of image.
        # Assuming that image is from the output of __getitem__
        rgb_image = np.clip(rgb_image, 0, 1)
        rgb_image = rgb_image ** (1 / gamma)

        rgb_image = np.clip(rgb_image, 0, 1)
        rgb_image *= 255
        rgb_image = rgb_image.astype("uint8")

        return rgb_image

    def get_image_paths_from_image_id(self, image_id):
        image_infos = self.coco_dset.index.imgs[image_id]["auxiliary"]

        data = {}
        for image_info in image_infos:
            band_name = image_info["channels"]
            rel_file_path = image_info["file_name"]

            data[band_name] = rel_file_path

        return data

    def get_pixel_normalization_params(self):
        Norm_Params = namedtuple("Norm_Params", ["mean", "std"])
        if "TA1" in self.dset_dir:
            norm_params_dict = {
                "B02": Norm_Params(mean=1.5640, std=1.5374),  # blue
                "B03": Norm_Params(mean=1.4271, std=1.4034),  # green
                "B04": Norm_Params(mean=1.5551, std=1.5293),  # red
                "B08": Norm_Params(mean=2.4479, std=2.4086),  # nir
            }
        else:
            norm_params_dict = {
                "B02": Norm_Params(mean=1.7603, std=1.7328),  # blue
                "B03": Norm_Params(mean=2.0488, std=2.0234),  # green
                "B04": Norm_Params(mean=1.9127, std=1.8875),  # red
                "B08": Norm_Params(mean=3.3785, std=3.3383),  # nir
            }
        mean_vals, std_vals = [], []
        if self.channels == "RGB":
            for band_name in ["B04", "B03", "B02"]:
                mean_vals.append(norm_params_dict[band_name].mean)
                std_vals.append(norm_params_dict[band_name].std)
        elif self.channels == "RGB_NIR":
            for band_name in ["B04", "B03", "B02", "B08"]:
                mean_vals.append(norm_params_dict[band_name].mean)
                std_vals.append(norm_params_dict[band_name].std)
        elif self.channels == "ALL":
            raise NotImplementedError("Global normalization parameters not generated for ALL channels.")

        mean_vals, std_vals = np.array(mean_vals), np.array(std_vals)

        # Format shapes.
        mean_vals = mean_vals[np.newaxis, :, np.newaxis, np.newaxis]
        std_vals = std_vals[np.newaxis, :, np.newaxis, np.newaxis]
        return mean_vals, std_vals

    def color_annotation(self, anno_image):
        """Create a RGB image containing annotations.

        Args:
            anno_image (np.array): A int numpy array of shape [height, width].

        Returns:
            np.array: A uint8 numpy arry of shape [height, width, 3].
        """
        height, width = anno_image.shape
        color_anno_image = np.zeros([height, width, 3], dtype="float")

        class_ids = np.unique(anno_image)
        for class_id in list(class_ids):
            if class_id not in [0, -1]:
                x, y = np.where(class_id == anno_image)
                rgb_value = self.cid_to_rgb(class_id)
                color_anno_image[x, y, :] = rgb_value

        color_anno_image = color_anno_image * 255
        color_anno_image = color_anno_image.astype("uint8")

        return color_anno_image

    def generate_propagated_annotation_gif(self, index, save_path=None):
        # Get video id.
        crop_info = self.examples[index]
        video_id = crop_info["video_id"]

        # Get image ids from video id.
        image_ids = self.video_dataset[video_id]

        # Get annotation images for each image ID.
        og_height, og_width = crop_info["og_height"], crop_info["og_width"]
        anno_images = [self.compute_segmentation_anno(img_id, og_height, og_width)[0] for img_id in image_ids]

        # Compute RGB versions of annotation images.
        rgb_anno_images = [self.color_annotation(a_img) for a_img in anno_images]

        # Generate GIF of annotation images.
        from watch.tasks.rutgers_material_change_detection.utils.util_misc import create_gif

        ## Get save path.
        if save_path is None:
            save_path = f"iarpa_anno_{str(index).zfill(3)}.gif"

        ## Get datetimes in str format.
        datetime_strs = [
            f"[{i}/{len(image_ids)}] " + self.get_datetime(img_id).strftime("%m/%d/%Y")
            for i, img_id in enumerate(image_ids)
        ]
        create_gif(rgb_anno_images, save_path, image_text=datetime_strs)


if __name__ == "__main__":
    from watch.tasks.rutgers_material_change_detection.utils.util_paths import get_dataset_root_dir
    from watch.tasks.rutgers_material_change_detection.utils.util_misc import generate_video_slice_object

    root_dir = get_dataset_root_dir("iarpa_drop1")
    # root_dir = get_dataset_root_dir('iarpa_drop1_ta1')
    task_mode = "total_bin_change"

    # Create dataset and load one example.
    split = "train"
    video_slice = generate_video_slice_object(height=150, width=150, n_frames=None, scale=2)
    dset = IARPA_KWDATASET(root_dir, split, video_slice, task_mode, channels="RGB_NIR", normalize_mode="local")
    example = dset.__getitem__(4)
    print("TEST 1: Load one example from dataset. | PASSED")

    # Check that validation set of dataset works.
    split = "valid"
    video_slice = generate_video_slice_object(height=400, width=400, n_frames=5, scale=None)
    dset = IARPA_KWDATASET(root_dir, split, video_slice, task_mode, channels="RGB_NIR")
    example = dset.__getitem__(0)
    print("TEST 2: Get an example from validation set. | PASSED")

    # Visualize one example of dataset.
    split = "valid"
    video_slice = generate_video_slice_object(height=300, width=300, n_frames=None, scale=None)
    dset = IARPA_KWDATASET(root_dir, split, video_slice, task_mode, channels="RGB_NIR")
    ex_index = 1
    dset.visualize_example(ex_index, save_path=f"iarpa_data_ex_{str(ex_index).zfill(2)}.png", overlay_last_anno=True)
    print("TEST 3: Create a visualizion. | PASSED")

    # Visualize datetimes of one example.
    # split = 'train'
    # video_slice = generate_video_slice_object(height=100, width=100, n_frames=None, scale=None)
    # dset = IARPA_KWDATASET(root_dir, split, video_slice, task_mode, channels='RGB_NIR')
    # ex_index = 10
    # dset.visualize_dates(ex_index, save_path=f'iarpa_data_dates_ex_{str(ex_index).zfill(2)}.png')
    # print('TEST 4: Create a dates visualizion. | PASSED')

    # Visualize image annotations over time.
    split = "valid"
    video_slice = generate_video_slice_object(height=800, width=800, n_frames=None, scale=None)
    dset = IARPA_KWDATASET(root_dir, split, video_slice, task_mode, channels="RGB_NIR")
    ex_index = 0
    dset.generate_propagated_annotation_gif(ex_index)
    print("TEST 5: Generate a video of annotations over a region. | PASSED")



# flake8: noqa
"""
This method uses K-Means clusters as region proposals,
which this tool allows you to label. Note that this tool:
    - Does not have "free brush" tool. It can only label clusters.
    - The processing time of each region proposal depends on the
    number of clusters, timesteps, and channels. It may take
    up to 2 minutes with some configurations.
"""

import sys
from PyQt5.QtGui import QPixmap, QImage  # NOQA
from PyQt5.QtWidgets import (QApplication, QWidget, QPushButton, QSizePolicy,   # NOQA
                             QDoubleSpinBox, QLabel, QCheckBox, QMainWindow,     # NOQA
                             QGridLayout, QComboBox, QTextBrowser, QLineEdit, QInputDialog)    # NOQA
from PyQt5.QtCore import Qt  # NOQA
from PyQt5 import QtCore
import cv2
import numpy as np
import matplotlib.pyplot as plt
import ubelt as ub
import kwcoco
import ndsampler
# import numpy as np
# import watch
# import matplotlib.pyplot as plt
# from tqdm import tqdm
import torch
# import pdb
from sklearn.cluster import KMeans
# from skimage.transform import resize
# from sklearn.manifold import TSNE
# import scipy
# from PIL import Image
# from PIL.ImageQt import ImageQt
import qimage2ndarray
import cmapy
# import random
import kwimage
from watch.tasks.rutgers_material_seg.datasets.iarpa_dataset import SequenceDataset


class Window(QMainWindow):
    def __init__(self, dataset, dset, resume='', save_path='', k=255):
        """Main interactive window

        Parameters
        ----------
        dataset : kwcoco
            KWCOCO dataset to use (with sampler)
        dset : kwcoco
            KWCOCO dataset to use (before sampler)
        resume : str, optional
            path to dataset to continue annotating, by default ''
        save_path : str, optional
            where to save the newly annotated dataset, by default ''
        k : int, optional
            number of clusters for K-Means algorithm, by default 255
        """
        super().__init__()
        self.image_counter = 0
        self.dataset = dataset
        self.channels, self.timesteps, self.im_width, self.im_height = self.dataset[self.image_counter]['inputs']['im'].data.shape
        self.dset = dset
        self.save_path = save_path
        self.k = k
        self.seen_labels, self.class_labels_all = [], []
        self.class_labels_pairs = {}
        self.class_label_with = "Concrete"
        self.class_label_to_index = {"Nothing": 0, "Concrete": 1, "Vegetation": 2, "Soil": 3, "Water": 4}
        if len(resume) > 1:
            self.material_dset = kwcoco.CocoDataset(resume)
            self.image_counter = int(len(list(self.material_dset.index.imgs.keys())) / self.timesteps) + 1
        else:
            self.material_dset = kwcoco.CocoDataset()
            for key, value in self.class_label_to_index.items():
                if key != 0:
                    self.material_dset.add_category(name=key, id=value)

        self.width, self.height = self.dataset[self.image_counter]['tr'].data['space_dims']

        self.scale_factor = 1
        self.scaled_height, self.scaled_width = int(self.height * self.scale_factor), int(self.width * self.scale_factor)
        self.current_mask = np.zeros((self.width, self.height)).astype(np.uint8)
        self.separable_current_mask = np.zeros((len(list(self.class_label_to_index.keys())), self.width, self.height)).astype(np.uint8)
        self.widget = QWidget()
        self.widget.keyPressEvent = self.keyPressEvent
        self.load_images(index=self.image_counter)

        self.label_img = QLabel(self.widget)
        self.image = QPixmap(self.qImg)
        self.label_img.setPixmap(self.image)
        self.label_img.setGeometry(20, 128, self.height, self.width)
        self.label_img.mousePressEvent = self.getImagePixel

        self.label_prediction = QLabel(self.widget)
        self.prediction = QPixmap(self.qPred)
        self.label_prediction.setPixmap(self.prediction)
        self.label_prediction.setGeometry(self.width + 20 + 120, 128, self.height, self.width)
        self.label_prediction.mousePressEvent = self.getMaskPixel

        self.label_mask = QLabel(self.widget)
        self.mask = QPixmap(self.qMask)
        self.label_mask.setPixmap(self.mask)
        self.label_mask.setGeometry(3 * self.width + 20, 128, self.height, self.width)

        self.output_textbox = QTextBrowser(self.widget)
        self.output_textbox.setGeometry(QtCore.QRect(20, self.height + 256 + 128 + 50, 100, 100))
        self.output_textbox.setObjectName("Current Clusters selected")

        self.output_textbox_name = QTextBrowser(self.widget)
        self.output_textbox_name.setGeometry(QtCore.QRect(140, self.height + 256 + 128 + 50, 100, 100))
        self.output_textbox_name.setObjectName("Labels of Selected Clusters")

        self.remove_cluster = QLineEdit(self.widget)
        self.remove_cluster.setObjectName("Cluster Removal Tool")
        self.remove_cluster.move(256, self.height + 256 + 128 + 50)
        self.remove_cluster.resize(100, 40)
        self.remove_cluster.textChanged.connect(self.textchanged)

        self.image_update = QPushButton(self.widget)
        self.image_update.setText("Previous Image")
        self.image_update.move(20, 20)
        self.image_update.clicked.connect(self.previous_image)

        self.image_update = QPushButton(self.widget)
        self.image_update.setText("Next Image")
        self.image_update.move(140, 20)
        self.image_update.clicked.connect(self.update_image)

        self.class_label = QComboBox(self.widget)
        self.class_label.addItems(["Nothing", "Concrete", "Vegetation", "Soil", "Water"])
        self.class_label.move(300, 20)
        self.class_label.currentIndexChanged.connect(self.class_label_selection)

        self.show_current_mask = QPushButton(self.widget)
        self.show_current_mask.setText("Show Current Mask")
        self.show_current_mask.move(500, 20)
        self.show_current_mask.clicked.connect(self.show_current_mask_clicked)

        self.increase_size = QPushButton(self.widget)
        self.increase_size.setText("Increase Image Size")
        self.increase_size.move(700, 20)
        self.increase_size.clicked.connect(self.increase_images_size)

        self.decrease_size = QPushButton(self.widget)
        self.decrease_size.setText("Decrease Image Size")
        self.decrease_size.move(900, 20)
        self.decrease_size.clicked.connect(self.decrease_images_size)

        self.timestamp_label = QComboBox(self.widget)
        self.timestamp_label.addItems([str(x) for x in range(self.timesteps)])
        self.timestamp_label.move(1100, 20)
        self.timestamp_label.currentIndexChanged.connect(self.change_inter_batch_image)

        self.gotoimage_label = QComboBox(self.widget)
        self.gotoimage_label.addItems([str(x) for x in range(self.dataset.__len__())])
        self.gotoimage_label.move(1200, 20)
        self.gotoimage_label.currentIndexChanged.connect(self.go_to_image)

        self.label_mask_title = QLabel(self.widget)
        self.label_mask_title.setText("Intermediate Mask")
        self.label_mask_title.move(2 * self.width + 256, 100)

        self.label_img_title = QLabel(self.widget)
        self.label_img_title.setText(f"Image {self.image_counter}")
        self.label_img_title.move(256, 100)

        self.label_prediction_title = QLabel(self.widget)
        self.label_prediction_title.setText("Clustering Prediction")
        self.label_prediction_title.move(self.width + 256, 100)

        self.remove_cluster_title = QLabel(self.widget)
        self.remove_cluster_title.setText("Write cluster to remove from the mask")
        self.remove_cluster_title.move(256, self.height + 256 + 128)

        self.output_textbox_title = QLabel(self.widget)
        self.output_textbox_title.setText("Current Clusters")
        self.output_textbox_title.move(20, self.height + 256 + 128)

        self.output_textbox_name_title = QLabel(self.widget)
        self.output_textbox_name_title.setText("Clusters Labels")
        self.output_textbox_name_title.move(140, self.height + 256 + 128)

        self.widget.setGeometry(50, 50, 1800, 800)
        self.widget.show()

    def change_inter_batch_image(self):
        """view a different image within the batch
        """
        image_data = self.dataset[self.image_counter]['inputs']['im'].data  # [b,c,t,h,w]
        image_data = image_data[:3, :, :self.width, :self.height]  # .copy()
        c, t, h, w = image_data.shape
        # random_t = random.randrange(t)
        self.t_selection = int(self.timestamp_label.currentText())
        # print(random_t)
        image_show = np.array(image_data[:, self.t_selection, :, :]).transpose(1, 2, 0).copy()  # visualize 0 indexe
        image_min = np.min(image_show)
        image_max = np.max(image_show)
        self.image_show = (image_show - image_min) / (image_max - image_min)
        self.qimage = qimage2ndarray.array2qimage(self.image_show * 255)  # .scaled(self.height, self.width)
        self.qImg = QPixmap(self.qimage)  # .scaled(256,256)

        self.image = QPixmap(self.qImg).scaled(self.scaled_height, self.scaled_width)
        self.label_img.setPixmap(self.image)
        # self.label_img.setGeometry(20,128,self.scaled_height, self.scaled_width)

    def increase_images_size(self):
        """enlarge size of displayed image and clusters
        """
        self.scale_factor += 0.25
        self.scaled_height, self.scaled_width = int(self.scale_factor * self.height), int(self.scale_factor * self.width)

        self.image = QPixmap(self.qImg).scaled(self.scaled_height, self.scaled_width)
        self.label_img.setPixmap(self.image)
        self.label_img.setGeometry(20, 128, self.scaled_height, self.scaled_width)

        self.prediction = QPixmap(self.qPred).scaled(self.scaled_height, self.scaled_width)
        self.label_prediction.setPixmap(self.prediction)
        self.label_prediction.setGeometry(self.scaled_width + 20 + 70 * (1 + self.scale_factor), 128, self.scaled_height, self.scaled_width)

        # self.mask = QPixmap(self.qMask).scaled(self.scaled_height, self.scaled_width)
        # self.label_mask.setPixmap(self.mask)
        self.label_mask.setGeometry(3 * self.scaled_width, 128, self.scaled_height, self.scaled_width)

    def decrease_images_size(self):
        """reduce size of displayed image and clusters
        """
        self.scale_factor -= 0.25
        self.scaled_height, self.scaled_width = int(self.scale_factor * self.height), int(self.scale_factor * self.width)

        self.image = QPixmap(self.qImg).scaled(self.scaled_height, self.scaled_width)
        self.label_img.setPixmap(self.image)
        self.label_img.setGeometry(20, 128, self.scaled_height, self.scaled_width)

        self.prediction = QPixmap(self.qPred).scaled(self.scaled_height, self.scaled_width)
        self.label_prediction.setPixmap(self.prediction)
        self.label_prediction.setGeometry(self.scaled_width + 20 + 50 * (1 + self.scale_factor), 128, self.scaled_height, self.scaled_width)

        # self.mask = QPixmap(self.qMask).scaled(self.scaled_height, self.scaled_width)
        # self.label_mask.setPixmap(self.mask)
        self.label_mask.setGeometry(3 * self.scaled_width, 128, self.scaled_height, self.scaled_width)

    def keyPressEvent(self, event):
        """allow change class to label with key press

        Parameters
        ----------
        event : int
            key press code
        """

        keys_to_class_dict = {48: "Nothing", 49: "Concrete", 50: "Vegetation", 51: "Soil", 52: "Water"}
        if event.key() in keys_to_class_dict.keys():
            self.class_label_with = keys_to_class_dict[event.key()]
            self.class_label.setCurrentText(keys_to_class_dict[event.key()])

    def show_current_mask_clicked(self):
        current_mask_no_bg = np.ma.masked_where(self.current_mask == 0, self.current_mask)
        plt.imshow(self.image_show)
        plt.imshow(current_mask_no_bg, alpha=0.6, cmap='tab20')
        plt.show()

    def finalize_mask(self):
        """Process current mask and add it to kwcoco dataset.
        """
        gids = self.dataset[self.image_counter]['tr'].data['gids']
        # widths_list = [x['width'] for x in self.dset.index.imgs[gids[0]]['auxiliary']]
        # max_dim_index = widths_list.index(max(widths_list))
        # max_dim_index = self.dset.index.imgs[gids[0]]['auxiliary'].index(max(self.dset.index.imgs[gids[0]]['auxiliary']))

        # im_space_width, im_space_height = self.dset.index.imgs[gids[0]]['auxiliary'][max_dim_index]['width'], self.dset.index.imgs[gids[0]]['auxiliary'][max_dim_index]['height']
        # transform = np.array([[1., 0, 0], [0, 1, 0], [0, 0, 1]])
        # type="polygon"
        for i in range(1, len(list(self.class_label_to_index.keys()))):
            if len(np.unique(self.separable_current_mask[i, :, :])) > 1:

                binary_mask = self.separable_current_mask[i, :, :]
                binary_mask = kwimage.Mask(binary_mask, format='c_mask')
                # binary_coco = binary_polygon.to_coco(style='new')
                # binary_segmentation = kwimage.Segmentation.coerce(binary_polygon)#.to_coco(style="new")
                for gid in gids:
                    image_dict =  self.dset.index.imgs[gid]
                    im_space_height, im_space_width = image_dict['height'], image_dict['width']  # NOQA
                    img_to_vid_transform = image_dict['warp_img_to_vid']['matrix']
                    img_to_vid_transform_npy = np.array(img_to_vid_transform)
                    img_to_vid_transform_inv_npy = np.linalg.inv(img_to_vid_transform_npy)

                    binary_polygon = binary_mask.to_multi_polygon()
                    mask = binary_polygon.warp(img_to_vid_transform_inv_npy)  # , output_dims=(im_space_height, im_space_width))
                    # binary_mask_2 = binary_mask.warp(img_to_vid_transform_inv_npy, output_dims=(im_space_height, im_space_width))
                    # binary_polygon_2 = binary_mask.to_multi_polygon()
                    self.material_dset.add_annotation(image_id=gid,
                                                      category_id=i,
                                                      bbox=ub.peek(mask.bounding_box().to_coco()),
                                                      segmentation=mask.to_coco(style="new")
                                                      # bbox=ub.peek(binary_polygon.bounding_box().to_coco()),
                                                      # segmentation=binary_polygon.to_coco(style="new")
                                                      )

        self.material_dset.validate()
        self.material_dset._check_integrity()
        self.material_dset.dump(self.save_path, newlines=True)

    def class_label_selection(self):
        """Select class to label.
        """
        self.class_label_with = self.class_label.currentText()

    def textchanged(self, text):
        """Remove a cluster
        """
        num, ok = QInputDialog.getInt(self, "Select a cluster to remove", "enter a number")
        if ok and num:
            if num in self.seen_labels:
                xs, ys = np.where(self.prediction_show == num)
                self.current_mask[xs, ys] = 0
                self.update_mask()
                self.seen_labels.remove(num)

    def getImagePixel(self, event):
        x = int(event.pos().x() // self.scale_factor)
        y = int(event.pos().y() // self.scale_factor)
        self.value = self.prediction_show[y, x]
        xs, ys = np.where(self.prediction_show == self.value)

        self.current_mask[xs, ys] = self.class_label_to_index[self.class_label_with]
        self.separable_current_mask[self.class_label_to_index[self.class_label_with], xs, ys] = 1  # need to account for removed pixels!

        non_label_indices = np.array(list(set(list(self.class_label_to_index.values())) - set([self.class_label_to_index[self.class_label_with]])))
        for label_index in non_label_indices:
            self.separable_current_mask[label_index, xs, ys] = 0

        self.update_mask()
        self.class_labels_pairs[self.value] = self.class_label_with
        self.seen_labels.append(self.value)
        self.output_textbox.append(str(self.value))
        self.output_textbox_name.append(str(self.class_label_with))

    def getMaskPixel(self, event):
        x = int(event.pos().x() // self.scale_factor)
        y = int(event.pos().y() // self.scale_factor)
        self.value = self.prediction_show[y, x]
        xs, ys = np.where(self.prediction_show == self.value)

        self.current_mask[xs, ys] = self.class_label_to_index[self.class_label_with]
        self.separable_current_mask[self.class_label_to_index[self.class_label_with], xs, ys] = 1
        non_label_indices = np.array(list(set(list(self.class_label_to_index.values())) - set([self.class_label_to_index[self.class_label_with]])))
        for label_index in non_label_indices:
            self.separable_current_mask[label_index, xs, ys] = 0

        self.update_mask()
        self.class_labels_pairs[self.value] = self.class_label_with
        self.seen_labels.append(self.value)
        self.output_textbox.append(str(self.value))
        self.output_textbox_name.append(str(self.class_label_with))

    def update_mask(self):
        self.current_mask_cmap = (self.current_mask * 20).astype(np.uint8)
        self.current_mask_cmap = cv2.applyColorMap(self.current_mask_cmap, cmapy.cmap('viridis'))
        self.qmask = qimage2ndarray.array2qimage(self.current_mask_cmap).scaled(self.height, self.width)
        self.mask = QPixmap(self.qmask)  # .scaled(256,256)
        # self.image = QPixmap(self.qImg)
        self.label_mask.setPixmap(self.mask)

    def previous_image(self):
        pass

    def go_to_image(self):
        plt.close()
        self.finalize_mask()
        self.class_labels_all.append(self.class_labels_pairs)
        self.class_labels_pairs = {}
        self.seen_labels = []
        self.output_textbox.clear()
        self.output_textbox_name.clear()
        self.image_counter = int(self.gotoimage_label.currentText())
        self.label_img_title.setText(f"Image {self.image_counter}")
        self.width, self.height = self.dataset[self.image_counter]['tr'].data['space_dims']
        self.scale_factor = 1
        self.current_mask = np.zeros((self.width, self.height)).astype(np.uint8)
        self.separable_current_mask = np.zeros((len(list(self.class_label_to_index.keys())), self.width, self.height)).astype(np.uint8)
        self.load_images(index=self.image_counter)
        # print("updated image and prediction")

        self.image = QPixmap(self.qImg)
        self.label_img.setPixmap(self.image)
        self.label_img.setGeometry(20, 128, self.height, self.width)

        self.prediction = QPixmap(self.qPred)
        self.label_prediction.setPixmap(self.prediction)
        self.label_prediction.setGeometry(self.width + 20 + 50, 128, self.height, self.width)

        self.mask = QPixmap(self.qMask)
        self.label_mask.setPixmap(self.mask)
        self.label_mask.setGeometry(2 * self.width + 128, 128, self.height, self.width)

    def update_image(self):
        plt.close()
        self.finalize_mask()
        self.class_labels_all.append(self.class_labels_pairs)
        # print(self.class_labels_all)
        self.class_labels_pairs = {}
        self.seen_labels = []
        self.output_textbox.clear()
        self.output_textbox_name.clear()
        self.image_counter += 1
        print(f"current iteration: {self.image_counter} \n width, height: {self.dataset[self.image_counter]['tr'].data['space_dims']}")
        self.label_img_title.setText(f"Image {str(self.image_counter)}")
        self.width, self.height = self.dataset[self.image_counter]['tr'].data['space_dims']
        self.scale_factor = 1
        self.current_mask = np.zeros((self.width, self.height)).astype(np.uint8)
        self.separable_current_mask = np.zeros((len(list(self.class_label_to_index.keys())), self.width, self.height)).astype(np.uint8)
        # self.width_factor, self.height_factor = self.vis_width/self.width, self.vis_height/self.height
        self.load_images(index=self.image_counter)
        # print("updated image and prediction")

        self.image = QPixmap(self.qImg)
        self.label_img.setPixmap(self.image)
        self.label_img.setGeometry(20, 128, self.height, self.width)

        self.prediction = QPixmap(self.qPred)
        self.label_prediction.setPixmap(self.prediction)
        self.label_prediction.setGeometry(self.width + 20 + 50, 128, self.height, self.width)

        self.mask = QPixmap(self.qMask)
        self.label_mask.setPixmap(self.mask)
        self.label_mask.setGeometry(2 * self.width + 128, 128, self.height, self.width)

    def load_images(self, index):
        kmeans = KMeans(n_clusters=self.k, random_state=0)
        image_data = self.dataset[index]['inputs']['im'].data  # [b,c,t,h,w]
        # print(f"image min:{image_data.min()} max: {image_data.max()}")
        gids = self.dataset[self.image_counter]['tr'].data['gids']
        for gid in gids:
            image_dict =  self.dset.index.imgs[gid]
            img_to_vid_transform = image_dict['warp_img_to_vid']['matrix']  # NOQA
            video_dict = self.dset.index.videos[image_dict['video_id']]
            if gid not in self.material_dset.index.imgs.keys():
                self.material_dset.add_image(**image_dict)

        if image_dict['video_id'] not in self.material_dset.index.videos.keys():
            self.material_dset.add_video(**video_dict)

        image_data = image_data[:, :, :self.width, :self.height]  # /20000#.copy()
        print(f"before image_data shape:{image_data.shape}")
        c, t, h, w = image_data.shape
        image_show = np.array(image_data[:3, 1, :, :]).transpose(1, 2, 0).copy()  # visualize 0 indexe
        image_min = np.min(image_show)
        image_max = np.max(image_show)
        self.image_show = (image_show - image_min) / (image_max - image_min)

        # print(f"image min:{self.image_show.min()} max: {self.image_show.max()}")
        # print(f"image counts: {np.unique(image_data, return_counts=True)}")
        # plt.imshow(self.image_show)
        # plt.show()

        image_data = image_data.contiguous().view(c, t, h * w)
        image_data = torch.transpose(image_data, 0, 2)
        image_data = torch.flatten(image_data, start_dim=1, end_dim=2)
        print(f"after image_data shape:{image_data.shape}")
        # print(f"image min:{image_data.min()} max: {image_data.max()}")
        kmeans.fit(image_data)
        cluster_labels = kmeans.labels_
        self.prediction_show = (cluster_labels.reshape(h, w)).astype(np.uint8)  # *6

        # self.fig = plt.figure()
        # ax = self.fig.add_subplot(1,1,1)
        # ax.imshow(self.prediction_show, cmap='viridis')
        # self.fig.show()

        self.prediction_show_cmap = cv2.applyColorMap(self.prediction_show, cmapy.cmap('viridis'))
        # print(f"prediciton shape: {self.prediction_show_cmap.shape}")
        b = self.prediction_show_cmap[:, :, 0]
        g = self.prediction_show_cmap[:, :, 1]
        r = self.prediction_show_cmap[:, :, 2]
        self.prediction_show_cmap = np.zeros(self.prediction_show_cmap.shape, dtype=np.uint8)
        self.prediction_show_cmap[:, :, 0] = r
        self.prediction_show_cmap[:, :, 1] = g
        self.prediction_show_cmap[:, :, 2] = b

        # self.fig = plt.figure()
        # ax = self.fig.add_subplot(1,1,1)
        # ax.imshow(self.prediction_show_cmap, cmap='viridis')
        # self.fig.show()

        # print(f"height: {self.height}, width: {self.width}")
        self.qimage = qimage2ndarray.array2qimage(self.image_show * 255)  # .scaled(self.height, self.width)
        self.qprediction = qimage2ndarray.array2qimage(self.prediction_show_cmap)  # .scaled(self.height, self.width)
        self.qmask = qimage2ndarray.array2qimage(self.current_mask)  # .scaled(self.height, self.width)

        self.qImg = QPixmap(self.qimage)  # .scaled(256,256)
        self.qPred = QPixmap(self.qprediction)  # .scaled(256,256)
        self.qMask = QPixmap(self.qmask)  # .scaled(256,256)


if __name__ == "__main__":

    app = QApplication(sys.argv)
    # save_kwcoco_path = "/home/native/core534_data/datasets/smart_watch/processed/drop0_aligned_v2.1/material_labels.kwcoco.json"
    resume = ""
    # resume = "/media/native/data/data/smart_watch_dvc/drop0_aligned_msi/material_labels2.kwcoco.json"
    save_kwcoco_path = "/media/native/data/data/smart_watch_dvc/drop0_aligned_msi/material_labels2.kwcoco.json"

    coco_fpath = ub.expandpath('/media/native/data/data/smart_watch_dvc/drop0_aligned_msi/data_fielded.kwcoco.json')
    # coco_fpath = ub.expandpath('/home/native/core534_data/datasets/smart_watch/processed/drop0_aligned_v2.1/data_fielded_filtered.kwcoco.json')

    # "cirrus", "coastal", "costal", "green", "lwir11", "lwir12", "nir", "pan", "red", "swir16", "swir22"]
    # channels = 'red|green|blue|nir|lwir11|lwir12|swir16|swir22'#|coastal|costal'
    channels = 'red|green|blue|nir|swir16|swir22|cirrus'
    # channels = 'red|green|blue|nir|swir22|cirrus'

    expected_channels = channels.split('|')
    print(expected_channels)
    # expected_channels = [ "red"  ,"green", "blue", "nir", "lwir11", "lwir12", "swir16", "swir22", "cirrus"]
    dset = kwcoco.CocoDataset(coco_fpath)

    # Only select images with the correct channels
    gids_to_remove = []
    for gid, img in dset.index.imgs.items():
        try:
            # print(img['auxiliary'][0]['channels'])
            image_auxiliary_list = img['auxiliary']
            bands_available = [x['channels'] for x in image_auxiliary_list]
            # print(bands_available)
            if not all(elem in bands_available for elem in expected_channels):
                gids_to_remove.append(gid)
        except Exception:
            gids_to_remove.append(gid)
            continue

    dset.remove_images(gids_to_remove)

    sampler = ndsampler.CocoSampler(dset)

    number_of_timestamps, h, w = 3, 512, 512
    window_dims = (number_of_timestamps, h, w)  # [t,h,w]
    input_dims = (h, w)

    dataset = SequenceDataset(sampler, window_dims, input_dims, channels)
    loader = dataset.make_loader(batch_size=1)

    window = Window(dataset, dset, resume, save_path=save_kwcoco_path)

    sys.exit(app.exec_())
